/*
 * delm_cmdu_ackq.c
 * CMDU response and ack queue management
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <libubus.h>
#include <libubox/utils.h>

#include <easy/easy.h>

#ifndef WFA_DELM_STANDALONE
#include <map-agent/plugin.h>
#include <map-agent/mapagent.h>
#include <map-agent/timer.h>
#else
#include "timer.h"
#endif /* WFA_DELM_STANDALONE */

#include "wfa_delm.h"
#include "delm.h"

static int timeradd_msecs(struct timeval *a, unsigned long msecs,
			  struct timeval *res)
{
	if (res) {
		struct timeval t = { 0 };

		if (msecs > 1000) {
			t.tv_sec += msecs / 1000;
			t.tv_usec = (msecs % 1000) * 1000;
		} else {
			t.tv_usec = msecs * 1000;
		}

		timeradd(a, &t, res);
		return 0;
	}

	return -1;
}

struct cmdu_ackq_entry *cmdu_ackq_create_msg(uint16_t type, uint16_t mid,
					     uint8_t *dest, uint32_t timeout)
{
	struct cmdu_ackq_entry *msg;
	struct timeval tsp = { 0 };

	msg = calloc(1, sizeof(*msg));
	if (!msg) {
		fprintf(stderr, "calloc failed. err = NOMEM\n");
		return NULL;
	}

	msg->type = type;
	msg->mid = mid;
	gettimeofday(&tsp, NULL);
	msg->ageing_time = timeout;
	timeradd_msecs(&tsp, msg->ageing_time, &msg->ageing_tmo);

	//msg->ageing_tmo.tv_usec = roundup(msg->ageing_tmo.tv_usec, 1000);
	msg->ageing_tmo.tv_usec = (msg->ageing_tmo.tv_usec / 1000) * 1000;
	memcpy(msg->origin, dest, 6);
	fprintf(stderr,
		"    CREATE msg: type = 0x%04x  mid = %hu origin = " MACFMT " timeout = { %u (%lu:%lu) }\n",
		type, mid, MAC2STR(dest), msg->ageing_time, msg->ageing_tmo.tv_sec, msg->ageing_tmo.tv_usec / 1000);

	return msg;
}

static void cmdu_ackq_delete_msg(struct cmdu_ackq_entry *msg)
{
	if (msg)
		free(msg);
}

static void cmdu_ackq_ageout_entry(struct cmdu_ackq *st, struct hlist_head *head,
				   struct timeval *min_next_tmo)
{
	struct delm_private *p = container_of(st,
			struct delm_private, cmdu_ack_q);
	struct wifi_network_device *dev = NULL;
	struct cmdu_ackq_entry *msg;
	struct hlist_node *tmp;
	struct timeval now = { 0 };


	gettimeofday(&now, NULL);
	now.tv_usec = (now.tv_usec / 1000) * 1000;


	hlist_for_each_entry_safe(msg, tmp, head, hlist) {
		fprintf(stderr, "%s(): Entry msg->ageout = (%lu.%lu), now = (%lu.%lu)\n",
				__func__,
				msg->ageing_tmo.tv_sec, msg->ageing_tmo.tv_usec,
				now.tv_sec, now.tv_usec);

		if (timercmp(&msg->ageing_tmo, &now, <=)) {
			st->pending_cnt--;
			hlist_del(&msg->hlist, head);
			fprintf(stderr, "No response from " MACFMT " for CMDU 0x%x with mid = %hu\n",
				MAC2STR(msg->origin), msg->type, msg->mid);

			/* mark the node as invalid node,
			 * reduce the pending count & update the datamodel
			 */
			p->pending--;
			dev = delm_get_origin_dev(p, msg->origin);
			if (dev) {
				/* TODO/REMOVE:
				 * mark the node as valid node to show the available data
				 */
				dev->is_valid_node = true;
				//fprintf(stderr, "marking the node as invalid node: "
				//		MACFMT "\n", MAC2STR(dev->macaddr));
				fprintf(stderr, "partial data has been received from this node: "
						MACFMT "\n", MAC2STR(dev->macaddr));
				if (p->pending <= 0)
					delm_update_datamodel(p);
			}

			cmdu_ackq_delete_msg(msg);
		} else {
			struct timeval new_next_tmo = { 0 };

			timersub(&msg->ageing_tmo, &now, &new_next_tmo);
			if (!timercmp(min_next_tmo, &new_next_tmo, <)) {
				min_next_tmo->tv_sec = new_next_tmo.tv_sec;
				min_next_tmo->tv_usec = new_next_tmo.tv_usec;
				/* fprintf("next-tmo = (%lu.%lu)\n",
				 *	   min_next_tmo->tv_sec,
				 *	   min_next_tmo->tv_usec);
				 */
			}
		}
	}
}

static void cmdu_ackq_ageing_timer_run(atimer_t *t)
{
	struct cmdu_ackq *st = container_of(t, struct cmdu_ackq, ageing_timer);
	//struct timeval *next_tmo = &st->next_tmo;
	struct timeval min_next_tmo = { .tv_sec = 999999 };
	int remain_cnt = 0;
	int i;
	struct timeval nu;

	gettimeofday(&nu, NULL);

	fprintf(stderr, "\n ----Timer now = %lu.%lu  --- cnt = %d---\n",
		nu.tv_sec, nu.tv_usec, st->pending_cnt);

	//spin_lock(&st->hash_lock);
	for (i = 0; i < CMDU_BACKLOG_MAX; i++) {
		if (hlist_empty(&st->table[i]))
			continue;

		fprintf(stderr, "i = %d\t", i);
		cmdu_ackq_ageout_entry(st, &st->table[i], &min_next_tmo);
	}

	remain_cnt = st->pending_cnt;
	st->next_tmo.tv_sec = min_next_tmo.tv_sec;
	st->next_tmo.tv_usec = min_next_tmo.tv_usec;

	//spin_unlock(&st->hash_lock);
	if (remain_cnt) {
		uint32_t tmo_msecs =
			st->next_tmo.tv_sec * 1000 + st->next_tmo.tv_usec / 1000;

		/* printf("%s(): remain = %d next_tmo = {(%lu s, %lu us) %u}\n",
		       __func__, remain_cnt, next_tmo->tv_sec, next_tmo->tv_usec, tmo_msecs); */

		if (tmo_msecs > 0)
			timer_set(&st->ageing_timer, tmo_msecs);
	}
}

int cmdu_ackq_init(void *cmdu_q)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;

	memset(q, 0, sizeof(*q));
	pthread_mutex_init(&q->qlock, NULL);
	timer_init(&q->ageing_timer, cmdu_ackq_ageing_timer_run);

	return 0;
}

struct cmdu_ackq_entry *cmdu_ackq_lookup(void *cmdu_q, uint16_t type,
					 uint16_t mid, uint8_t *dest)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	int idx = cmdu_ackq_hash(type, mid, dest);
	struct cmdu_ackq_entry *msg = NULL;

	pthread_mutex_lock(&q->qlock);
	hlist_for_each_entry(msg, &q->table[idx], hlist) {
		if (msg->type == type && msg->mid == mid &&
		    !memcmp(msg->origin, dest, 6)) {

			pthread_mutex_unlock(&q->qlock);
			return msg;
		}
	}
	pthread_mutex_unlock(&q->qlock);

	return NULL;
}

void cmdu_ackq_reset(void *cmdu_q)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	struct cmdu_ackq_entry *msg = NULL;
	int idx = 0;

	fprintf(stderr, "Resetting CMDU Message Queue..\n");
	fprintf(stderr, "*****************************************\n");
	pthread_mutex_lock(&q->qlock);
	for (idx = 0; idx < CMDU_BACKLOG_MAX; idx++) {
		hlist_for_each_entry(msg, &q->table[idx], hlist) {
			fprintf(stderr, "idx:%d, type:%d, mid:%d\n", idx, msg->type, msg->mid);
			cmdu_ackq_delete_msg(msg);
		}
		q->table[idx].first = NULL;
	}
	fprintf(stderr, "*****************************************\n");

	q->pending_cnt = 0;
	pthread_mutex_unlock(&q->qlock);
}

void cmdu_ackq_free(void *cmdu_q)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;

	cmdu_ackq_reset(q);

	timer_del(&q->ageing_timer);
	pthread_mutex_destroy(&q->qlock);
}

/* In this function, type = cmdutype that is expected with 'mid' from 'dest' */
int cmdu_ackq_enqueue(void *cmdu_q, uint16_t type, uint16_t mid, uint8_t *dest,
		      uint32_t timeout)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	struct cmdu_ackq_entry *msg = NULL;

	msg = cmdu_ackq_lookup(cmdu_q, type, mid, dest);
	if (msg) {
		fprintf(stderr,
			"Duplicate: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
			type, mid, MAC2STR(dest));

		return -1;
	}

	msg = cmdu_ackq_create_msg(type, mid, dest, timeout);
	if (msg) {
		int idx = cmdu_ackq_hash(type, mid, dest);

		pthread_mutex_lock(&q->qlock);
		hlist_add_head(&msg->hlist, &q->table[idx]);

		q->pending_cnt++;
		fprintf(stderr,
			"ENQ: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
			type, mid, MAC2STR(dest));


		if (timer_pending(&q->ageing_timer)) {
			if (timercmp(&q->next_tmo, &msg->ageing_tmo, >)) {
				q->next_tmo.tv_sec = msg->ageing_tmo.tv_sec;
				q->next_tmo.tv_usec = msg->ageing_tmo.tv_usec;

				timer_set(&q->ageing_timer, msg->ageing_time);
				fprintf(stderr, "Adjust ageout timer ========>\n");
			}
		} else {
			fprintf(stderr, "Start ageout timer ========>\n");
			q->next_tmo.tv_sec = msg->ageing_tmo.tv_sec;
			q->next_tmo.tv_usec = msg->ageing_tmo.tv_usec;
			timer_set(&q->ageing_timer, msg->ageing_time);
		}

		pthread_mutex_unlock(&q->qlock);
		return 0;
	}

	return -1;
}

int cmdu_ackq_dequeue(void *cmdu_q, uint16_t type, uint16_t mid, uint8_t *src)
{
	struct cmdu_ackq *q = (struct cmdu_ackq *)cmdu_q;
	struct cmdu_ackq_entry *msg = NULL;
	int idx;

	msg = cmdu_ackq_lookup(cmdu_q, type, mid, src);
	if (!msg) {
		fprintf(stderr,
			"DROP! CMDU not found: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
			type, mid, MAC2STR(src));

		return -1;
	}

	idx = cmdu_ackq_hash(type, mid, src);

	pthread_mutex_lock(&q->qlock);
	hlist_del(&msg->hlist, &q->table[idx]);
	q->pending_cnt--;
	pthread_mutex_unlock(&q->qlock);

	fprintf(stderr,
		"DEQ: type = 0x%04x  mid = %hu origin = " MACFMT "\n",
		type, mid, MAC2STR(src));

	cmdu_ackq_delete_msg(msg);

	return 0;
}

int delm_update_stalog_data(void *sta_log, const char *ifname,
				uint8_t *mac)
{
	int ret;
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *ev = NULL;

	if (!slog->active_sta_cnt)
		return 0;

	pthread_mutex_lock(&slog->evlock);
	hlist_for_each_entry(ev, &slog->start, hlist) {
		if ((!memcmp(ev->mac, mac, 6)) && (!strcmp(ev->ifname, ifname))) {
			ret = wifi_get_sta_info(ev->ifname, ev->mac, &ev->sta);
			if (ret) {
				fprintf(stderr, "Failed to fetch STA info for " MACFMT "\n",
						MAC2STR(ev->mac));
				break;
			}
		}
	}
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

void refresh_sta_timer_cb(atimer_t *t)
{
	struct sta_eventlog_entry *msg = container_of(t, struct sta_eventlog_entry, t);
	struct sta_eventlog *sv = msg->sv;

	delm_update_stalog_data(sv, msg->ifname, msg->mac);

	timer_set(t, 1 * 1000);
}

int sta_eventlog_init(void *sta_log)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;

	memset(slog, 0, sizeof(*slog));
	pthread_mutex_init(&slog->evlock, NULL);
	return 0;
}

struct sta_eventlog_entry *sta_eventlog_lookup(void *sta_log, uint8_t *mac)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *msg = NULL;

	pthread_mutex_lock(&slog->evlock);
	hlist_for_each_entry(msg, &slog->start, hlist) {
		if (!memcmp(msg->mac, mac, 6)) {
			pthread_mutex_unlock(&slog->evlock);
			return msg;
		}
	}
	pthread_mutex_unlock(&slog->evlock);
	return NULL;
}

int sta_eventlog_enqueue(void *sta_log, const char *ifname, uint8_t *mac)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *msg = NULL;

	msg = sta_eventlog_lookup(sta_log, mac);
	if (msg) {
		fprintf(stderr, "Duplicate STA event: macaddr = " MACFMT "\n",
				MAC2STR(mac));
		return -1;
	}

	msg = calloc(1, sizeof(*msg));
	if (!msg) {
		fprintf(stderr, "calloc failed. err = NOMEM\n");
		return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	strncpy(msg->ifname, ifname, 15);
	memcpy(msg->mac, mac, 6);
	timer_init(&msg->t, refresh_sta_timer_cb);
	timer_set(&msg->t, 1 * 1000);
	msg->sv = slog;
	hlist_add_head(&msg->hlist, &slog->start);
	slog->active_sta_cnt++;
	fprintf(stderr, "STA ENQ: macaddr = " MACFMT "\n", MAC2STR(mac));
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

int sta_eventlog_dequeue(void *sta_log, uint8_t *mac)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *msg = NULL;

	msg = sta_eventlog_lookup(sta_log, mac);
	if (!msg) {
		fprintf(stderr, "DROP! STA not found: macaddr = " MACFMT "\n",
				MAC2STR(mac));
		return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	timer_del(&msg->t);
	hlist_del(&msg->hlist, &slog->start);
	slog->active_sta_cnt--;
	fprintf(stderr, "STA DEQ: macaddr = " MACFMT "\n", MAC2STR(mac));
	free(msg);
	msg = NULL;
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}
