/*
 * delm.c
 * WiFi Data Elements plugin
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubus.h>
#include <libubox/utils.h>

#include <easy/easy.h>

#ifndef WFA_DELM_STANDALONE
#include <map-agent/plugin.h>
#include <map-agent/mapagent.h>
#include <map-agent/timer.h>
#else
#include "timer.h"
#endif /* WFA_DELM_STANDALONE */

#include "wfa_delm.h"
#include "delm.h"



void free_dev_data(struct wifi_network_device *dev)
{
	int i, j, k, l;
	struct wifi_radio_element *radio = NULL;
	struct wifi_bss_element *bss = NULL;
	struct wifi_scanres_element *scanres = NULL;
	struct wifi_scanres_opclass_element *op = NULL;
	struct wifi_scanres_channel_element *ch = NULL;

	for (i = 0; i < dev->num_radios; i++) {
		radio = dev->radiolist + i;

		if (radio->num_supp_opclass > 0)
			free(radio->supp_opclass);

		radio->num_supp_opclass = 0;
		radio->supp_opclass = NULL;

		if (radio->num_curr_opclass > 0)
			free(radio->curr_opclass);

		radio->num_curr_opclass = 0;
		radio->curr_opclass = NULL;

		if (radio->num_unassoc_sta > 0)
			free(radio->unassoc_stalist);

		radio->num_unassoc_sta = 0;
		radio->unassoc_stalist = NULL;

		for (j = 0; j < radio->num_bss; j++) {
			bss = radio->bsslist + j;
			if (bss->num_stations > 0)
				free(bss->stalist);

			bss->num_stations = 0;
			bss->stalist = NULL;
		}

		if (radio->num_bss)
			free(radio->bsslist);

		radio->num_bss = 0;
		radio->bsslist = NULL;

		for (j = 0; j < radio->num_scanresult; j++) {
			scanres = radio->scanlist + j;
			if (scanres->num_opclass_scanned > 0) {
				for (k = 0; k < scanres->num_opclass_scanned; k++) {
					op = scanres->opclass_scanlist + k;
					if (op->num_channels_scanned > 0) {
						for (l = 0; l < op->num_channels_scanned; l++) {
							ch = op->channel_scanlist + l;
							if (ch->num_neighbors > 0)
								free(ch->nbrlist);
						}
						free(op->channel_scanlist);
					}
				}
				free(scanres->opclass_scanlist);
			}
		}

		if (radio->num_scanresult > 0)
			free(radio->scanlist);

		radio->num_scanresult = 0;
		radio->scanlist = NULL;
	}

	if (dev->num_radios)
		free(dev->radiolist);

	dev->radiolist = NULL;
	dev->num_radios = 0;

	if (dev->ieee1905_tlv_data)
		free(dev->ieee1905_tlv_data);

	dev->ieee1905_tlv_data = NULL;
}

void refresh_timer_cb(atimer_t *t)
{
	struct delm_private *p = container_of(t, struct delm_private, t);

	fprintf(stderr, "wfa_delm: %s: periodic = '%d' ms.\n",
		__func__, p->cfg.refresh_int);

	if (!p->getter_running) {
		if (p->cfg.is_collector)
			delm_collector_run(p);
		else
			delm_getter_run(p);

		goto rearm_refresh_timer;
	}

rearm_refresh_timer:
	if (p->cfg.refresh_int)
		timer_set(t, p->cfg.refresh_int);
}

int delm_exit(void *priv)
{
	int i;
	struct wifi_network *net;
	struct wifi_network_device *dev;
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct wifi_sta_events *sv = NULL;
	struct delm_private *p = (struct delm_private *)priv;

	if (p) {
		ubus_unregister_event_handler(p->ctx, &p->ieee1905_cmdu);

		if (!p->cfg.is_collector) {
			ubus_unregister_event_handler(p->ctx, &p->ieee1905_ev);
			ubus_unregister_event_handler(p->ctx, &p->sta_ev);
		}

		if (p->cfg.is_collector)
			ubus_unregister_event_handler(p->ctx, &p->topology_ev);

		if (p->dm) {
			net = &p->dm->network;
			if (net->num_devices > 0) {
				for (i = 0; i < net->num_devices; i++) {
					dev = net->device + i;
					free_dev_data(dev);
				}
			}
			free(p->dm);
		}

		obj = &p->obj;
		obj_methods = (struct ubus_method *)obj->methods;
		if (obj_methods)
			free(obj_methods);

		obj_type = obj->type;
		if (obj_type)
			free(obj_type);

		ubus_free(p->ctx);

		sv = p->sv;
		if (sv) {
			if (sv->num_assoc_events > 0)
				free(sv->assoclist);

			if (sv->num_disassoc_events > 0)
				free(sv->disassoclist);

			free(sv);
		}

		cmdu_ackq_free(&p->cmdu_ack_q);
		free(p);
	}

	return 0;
}

int delm_init(void **priv, void *cfg)
{
	struct delm_private *p;
	int ret;

	*priv = NULL;
	p = calloc(1, sizeof(struct delm_private));
	if (!p)
		return -1;

	cmdu_ackq_init(&p->cmdu_ack_q);
	p->dm = NULL;
	p->dmi = NULL;

	p->sv = calloc(1, sizeof(struct wifi_sta_events));
	if (!p->sv)
		goto out_err;

	if (cfg) {
		struct delm_config *c = (struct delm_config *)cfg;

		p->cfg.uctx = c->uctx;
		p->cfg.refresh_int = c->refresh_int;
		p->cfg.is_collector = c->is_collector;
		p->cfg.is_sap_mode = c->is_sap_mode;
		p->cfg.profile = c->profile;

		p->ctx = c->uctx;
	}



	fprintf(stderr, "%s:  priv = %p  (collector = %d)\n",
			__func__, p, p->cfg.is_collector);

	ret = delm_publish_object(p, p->cfg.is_collector ?
				  WFADATAD_COLLECTOR_OBJECT :
				  WFADATAD_AGENT_OBJECT);
	if (ret)
		goto out_err;

	p->ieee1905 = delm_lookup_object(p, "ieee1905");
	if (p->ieee1905 == OBJECT_INVALID) {
		err("Failed to get 'ieee1905' object!\n");
		goto out_err;
	}

	p->map1905 = delm_lookup_object(p, "map.1905");
	if (p->map1905 == OBJECT_INVALID) {
		err("Failed to get 'map.1905' object!\n");
		goto out_err;
	}

	p->topology = delm_lookup_object(p, "topology");
	if (p->topology == OBJECT_INVALID) {
		err("Failed to get 'topology' object!\n");
		goto out_err;
	}

	ret = delm_get_ieee1905_interfaces(p);
	if (ret)
		goto out_err;


	ret = delm_register_events(p);
	if (ret)
		goto out_err;

	if (!p->cfg.is_sap_mode) {
		/* send AP-autoconfig search message to get the controller id */
		ret = prepare_autoconfig_search(p, 0x00);	/* 2.4 GHZ */
		if (ret < 0)
			fprintf(stderr, "error in sending ap-autoconfig message\n");
	}


	/* start the DE-collector or DE-agent */
	if (p->cfg.is_collector)
		delm_collector_run(p);
	else
		delm_getter_run(p);


	timer_init(&p->t, refresh_timer_cb);
	if (p->cfg.refresh_int)
		timer_set(&p->t, p->cfg.refresh_int);

	/* self node topology query */
	if (!p->cfg.is_collector)
		self_node_query(p);

	*priv = p;
	return 0;

out_err:
	cmdu_ackq_free(&p->cmdu_ack_q);
	free(p);
	return -1;
}


#ifndef WFA_DELM_STANDALONE
extern struct plugin wfa_delm;
struct plugin wfa_delm = {
	.id = "\x41\x2e\x63\x78",
	.name = "wfa_delm",
	.init = delm_init,
	.exit = delm_exit,
};
#endif
