/*
 * delm_cmdu.c
 * File contains CMDUs handling needed for WiFi Data Elements.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: saurabh.verma@iopsys.eu
 *	   anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <net/if_arp.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <net/if.h>

#include <easy/easy.h>

#ifndef WFA_DELM_STANDALONE
#include <map-agent/mapagent.h>
#include <map-agent/timer.h>
#else
#include "timer.h"
#endif

#include <map1905/map2.h>
#include <map1905/maputils.h>

#include "wfa_delm.h"
#include "delm.h"


#define UBUS_MAP1905		"map.1905"
#define UBUS_IEEE1905		"ieee1905"
#define UBUS_TOPOLOGY		"topology"
#define UBUS_TIMEOUT            1000
#define MAX_INTF_PER_RADIO	16
#define INTERFACE_LEN		16
#define AP_COLLECTION_INTERVAL	(10 * 1000)


uint8_t *extract_tlv_by_type(struct cmdu_cstruct *cmdu, uint8_t tlv_type)
{
	uint8_t *tlv;
	int i;

	for (i = 0; i < cmdu->num_tlvs; i++) {
		tlv = cmdu->tlvs[i];
		if (*tlv == tlv_type)
			return tlv;
	}

	return NULL;
}

/* Fill the data elements using libwifi API's */
static void get_supp_services(struct wifi_network_device *own,
					struct tlv_supp_service *p)
{
	int i;

	p->supported_services_list = 1;
	if (p->supported_services_list > 0)
		p->supported_services = calloc(p->supported_services_list, sizeof(*p->supported_services));

	if (p->supported_services) {
		for (i = 0; i < p->supported_services_list; i++)
			p->supported_services[i].service = SUPPORTED_SERVICE_MULTIAP_AGENT;
	}
}

static void fill_ap_operational_bss(struct wifi_network_device *own,
						struct tlv_ap_oper_bss *p)
{
	uint32_t i, j;

	p->radios_nr = own->num_radios;
	p->radio = calloc(p->radios_nr, sizeof(*p->radio));

	if (!p->radio) {
		p->radios_nr = 0;
		return;
	}

	for (i = 0; i < p->radios_nr; i++) {
		memcpy(p->radio[i].radio_id, own->radiolist[i].macaddr, 6);
		p->radio[i].bss_nr = own->radiolist[i].num_bss;
		p->radio[i].bss = calloc(p->radio[i].bss_nr, sizeof(*(p->radio[i].bss)));

		if (!p->radio[i].bss) {
			p->radio[i].bss_nr = 0;
			return;
		}

		for (j = 0; j < p->radio[i].bss_nr; j++) {
			memcpy(p->radio[i].bss[j].bssid, own->radiolist[i].bsslist[j].bssid, 6);
			strcpy(p->radio[i].bss[j].ssid, own->radiolist[i].bsslist[j].ssid);
			p->radio[i].bss[j].ssid_len = strlen(p->radio[i].bss[j].ssid);
		}
	}
}

static void get_associated_clients(struct wifi_network_device *own,
					struct tlv_assoc_client *p)
{
	uint32_t i, j, k;
	uint32_t start, end;

	for (i = 0; i < own->num_radios; i++) {
		start = p->bss_nr;
		struct wifi_radio_element *radio = own->radiolist + i;
		p->bss_nr += radio->num_bss;
		end = p->bss_nr;
		if (start < end) {
			p->bss = realloc(p->bss, p->bss_nr * sizeof(*p->bss));
			if (p->bss) {
				for (j = start; j < end; j++) {
					memset(&p->bss[j], 0, sizeof(*p->bss));
					struct wifi_bss_element *bss = radio->bsslist + j - start;
					memcpy(p->bss[j].bssid, bss->bssid, 6);
					p->bss[j].assoc_clients_nr = bss->num_stations;
					if (p->bss[j].assoc_clients_nr > 0) {
						p->bss[j].clients = calloc(p->bss[j].assoc_clients_nr,
								sizeof(*(p->bss[j].clients)));
						if (!p->bss[j].clients) {
							p->bss[j].assoc_clients_nr = 0;
							continue;
						}
						for (k = 0; k < p->bss[j].assoc_clients_nr; k++) {
							memcpy(p->bss[j].clients[k].client_addr,
									bss->stalist[k].macaddr, 6);
							p->bss[j].clients[k].uptime = bss->stalist[k].conn_time;
						}
					}
				}
			}
		}
	}
}

static void get_multiap_profile(struct wifi_network_device *own,
					struct tlv_map_profile *p)
{
	p->profile = MULTIAP_PROFILE_2;
}

static void get_ap_radio_basic_cap(struct wifi_network_device *own,
					struct tlv_ap_radio_basic_cap *p,
					uint32_t radio_index)
{
	uint32_t j;
	struct wifi_radio_element *radio = own->radiolist + radio_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	p->max_bss_nr = 4;	// TODO: Max BSS per Radio
	p->operating_classes_nr = radio->num_supp_opclass;

	if (p->operating_classes_nr > 0)
		p->operating_class = calloc(p->operating_classes_nr, sizeof(*p->operating_class));
	if (!p->operating_class) {
		fprintf(stderr, "|%s:%d| out of memory!\n", __func__, __LINE__);
		p->operating_classes_nr = 0;
		return;
	}

	for (j = 0; j < p->operating_classes_nr; j++) {
		p->operating_class[j].op_class = radio->supp_opclass[j].id;
		p->operating_class[j].max_tx_power = radio->supp_opclass[j].max_txpower;
		p->operating_class[j].non_op_ch_nr = radio->supp_opclass[j].num_exclude_channels;
		if (p->operating_class[j].non_op_ch_nr > 0) {
			p->operating_class[j].channel =
				(uint8_t *)calloc(p->operating_class[j].non_op_ch_nr, sizeof(uint8_t));
			if (p->operating_class[j].channel) {
					memcpy(p->operating_class[j].channel, radio->supp_opclass[j].exclude_chanlist,
							p->operating_class[j].non_op_ch_nr);
			}
		}
	}
}

static void get_collection_interval(struct wifi_network_device *own,
				struct tlv_metric_collection_interval *p)
{
	p->collection_interval = AP_COLLECTION_INTERVAL;
}

static void get_ap_ht_caps(struct wifi_network_device *own,
				struct tlv_ap_ht_cap *p, uint32_t radio_index)
{
	struct wifi_radio_element *radio;

	radio = own->radiolist + radio_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	p->max_tx_streams_supported = (radio->caps.ht & 0xc0) >> 6;
	p->max_rx_streams_supported = (radio->caps.ht & 0x30) >> 4;
	p->gi_20_support = radio->caps.ht & (1 << 3) ? 1 : 0;
	p->gi_40_support = radio->caps.ht & (1 << 2) ? 1 : 0;
	p->ht_40_support = radio->caps.ht & (1 << 1) ? 1 : 0;
}

static void get_ap_vht_caps(struct wifi_network_device *own,
				struct tlv_ap_vht_cap *p, uint32_t radio_index)
{
	struct wifi_radio_element *radio = own->radiolist + radio_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	memcpy(&p->vht_tx_mcs_supported, &radio->caps.vht[0], 2);
	memcpy(&p->vht_rx_mcs_supported, &radio->caps.vht[2], 2);
	p->max_tx_streams_supported = (radio->caps.vht[4] & 0xe0) >> 5;
	p->max_rx_streams_supported = (radio->caps.vht[4] & 0x1c) >> 3;

	p->gi_80_support = radio->caps.vht[4] & (1 << 1) ? 1 : 0;
	p->gi_160_support = radio->caps.vht[4] & (1 << 0) ? 1 : 0;
	p->vht_8080_support = radio->caps.vht[5] & (1 << 7) ? 1 : 0;
	p->vht_160_support = radio->caps.vht[5] & (1 << 6) ? 1 : 0;
	p->su_beamformer_capable = radio->caps.vht[5] & (1 << 5) ? 1 : 0;
	p->mu_beamformer_capable = radio->caps.vht[5] & (1 << 4) ? 1 : 0;
}

static void get_ch_scan_caps(struct wifi_network_device *own,
					struct tlv_ch_scan_cap *p)
{
	int i, j;
	struct wifi_radio_element *radio = NULL;
	struct wifi_opclass_supported_element *opclass = NULL;

	p->nbr_radios = own->num_radios;

	if (p->nbr_radios > 0)
		p->radio = calloc(p->nbr_radios, sizeof(*p->radio));

	if (p->radio) {
		for (i = 0; i < p->nbr_radios; i++) {
			radio = own->radiolist + i;
			memcpy(p->radio[i].radio_id, radio->macaddr, 6);
			p->radio[i].on_boot = 0;
			p->radio[i].scan_impact = 0x03;
			p->radio[i].min_scan_interval = 30;
			p->radio[i].nbr_op_classes = radio->num_supp_opclass;

			if (p->radio[i].nbr_op_classes > 0) {
				p->radio[i].op_class = calloc(p->radio[i].nbr_op_classes,
								sizeof(*p->radio[i].op_class));
			}

			if (p->radio[i].op_class) {
				for (j = 0 ; j < radio->num_supp_opclass; j++) {
					opclass = radio->supp_opclass + j;
					p->radio[i].op_class[j].op_class = opclass->id;
					p->radio[i].op_class[j].nbr_ch = 0; // Scan on all channels

					if (p->radio[i].op_class[j].nbr_ch > 0) {
						p->radio[i].op_class[j].ch = calloc(p->radio[i].op_class[j].nbr_ch,
										sizeof(*p->radio[i].op_class[j].ch));
					}
				}
			}
		}
	}
}

static void get_cac_caps(struct wifi_network_device *own,
					struct tlv_cac_cap *p)
{
	int i, j, k, l;
	uint8_t ch;
	struct wifi_radio_element *radio = NULL;
	struct wifi_opclass_supported_element *opclass = NULL;

	memcpy(p->country_code, own->radiolist[0].country_code, 2);
	p->nbr_radios = own->num_radios;

	if (p->nbr_radios > 0)
		p->radio_data = calloc(p->nbr_radios, sizeof(*p->radio_data));

	if (!p->radio_data) {
		p->nbr_radios = 0;
		return;
	}

	for (i = 0; i < p->nbr_radios; i++) {
		radio = own->radiolist + i;
		memcpy(p->radio_data[i].radio_id, radio->macaddr, 6);
		p->radio_data[i].nbr_cac_types = 1;

		if (p->radio_data[i].nbr_cac_types > 0) {
			p->radio_data[i].cac_data = calloc(p->radio_data[i].nbr_cac_types,
							sizeof(*p->radio_data[i].cac_data));

			if (!p->radio_data[i].cac_data) {
				p->radio_data[i].nbr_cac_types = 0;
				continue;
			}

			for (j = 0; j < p->radio_data[i].nbr_cac_types; j++) {
				p->radio_data[i].cac_data[j].cac_method_supported = CAC_CAP_METHOD_CONTI_CAC;
				p->radio_data[i].cac_data[j].duration = 60; //Default duration is 60, 600 for channels (120, 124 & 128)
				// TODO: Assumption, All the supporting op_class supported continuous CAC method,
				// Currently no parmaeter/method to classify the same.
				p->radio_data[i].cac_data[j].nbr_op_classes = radio->num_supp_opclass;

				if (p->radio_data[i].cac_data[j].nbr_op_classes > 0) {
					p->radio_data[i].cac_data[j].op_class_data =
						calloc(radio->num_supp_opclass,
								sizeof(*p->radio_data[i].cac_data[j].op_class_data));
				}

				if (p->radio_data[i].cac_data[j].op_class_data) {
					for (k = 0; k < radio->num_supp_opclass; k++) {
						opclass = radio->supp_opclass + k;
						p->radio_data[i].cac_data[j].op_class_data[k].op_class = opclass->id;
						p->radio_data[i].cac_data[j].op_class_data[k].nbr_ch = opclass->num_channels;

						if (opclass->num_channels > 0) {
							p->radio_data[i].cac_data[j].op_class_data[k].ch =
								calloc(opclass->num_channels, sizeof(uint8_t));
						}

						if (p->radio_data[i].cac_data[j].op_class_data[k].ch) {
							for (l = 0; l < opclass->num_channels; l++) {
								p->radio_data[i].cac_data[j].op_class_data[k].ch[l] = opclass->chanlist[l];

								ch = p->radio_data[i].cac_data[j].op_class_data[k].ch[l];
								if (ch == 120 || ch == 124 || ch == 128)
									p->radio_data[i].cac_data[j].duration = 600;
							}
						}
					}
				}
			}
		}
	}
}

static void get_profile2_caps(struct wifi_network_device *own,
				struct tlv_profile2_ap_cap *p)
{
	p->byte_counter = 0;
	p->max_vids = (uint8_t)255;
}

static void get_radio_metrics(struct wifi_network_device *own,
				struct tlv_radio_metrics *p,
				uint32_t radio_index)
{
	struct wifi_radio_element *radio = own->radiolist + radio_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	p->noise = radio->anpi;
	p->transmit = radio->tx_utilization;
	p->receive_self = radio->rx_utilization;
	p->receive_other = radio->other_utilization ;
}

static void get_ap_metrics(struct wifi_network_device *own,
				struct tlv_ap_metrics *p, uint32_t radio_index,
				uint32_t bss_index)
{
	struct wifi_radio_element *radio = own->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;

	memcpy(p->bssid, bss->bssid, 6);
	if (bss->is_ac_be) {
		p->is_ac_be = bss->is_ac_be;
		memcpy(p->service_param_info_be, bss->est_wmm_be, 3);
	}
	if (bss->is_ac_bk) {
		p->is_ac_bk = bss->is_ac_bk;
		memcpy(p->service_param_info_bk, bss->est_wmm_bk, 3);
	}
	if (bss->is_ac_vo) {
		p->is_ac_vo = bss->is_ac_vo;
		memcpy(p->service_param_info_vo, bss->est_wmm_vo, 3);
	}
	if (bss->is_ac_vi) {
		p->is_ac_vi = bss->is_ac_vi;
		memcpy(p->service_param_info_vi, bss->est_wmm_vi, 3);
	}
}

static void get_ap_ext_metrics(struct wifi_network_device *own,
				struct tlv_ap_ext_metrics *p,
				uint32_t radio_index, uint32_t bss_index)
{
	struct wifi_radio_element *radio = own->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;

	memcpy(p->bssid, bss->bssid, 6);
	p->uni_bytes_sent = bss->tx_ucast_bytes;
	p->uni_bytes_recv = bss->rx_ucast_bytes;
	p->multi_bytes_sent = bss->tx_mcast_bytes;
	p->multi_bytes_recv = bss->rx_mcast_bytes;
	p->bro_bytes_sent = bss->tx_bcast_bytes;
	p->bro_bytes_recv = bss->rx_bcast_bytes;
}

static void get_sta_link_metrics(struct wifi_network_device *own,
				struct tlv_assoc_sta_link_metrics *p,
				uint32_t radio_index,
				uint32_t bss_index, uint32_t sta_index)
{
	int i;
	struct wifi_radio_element *radio = own->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;
	struct wifi_sta_element *sta = bss->stalist + sta_index;

	memcpy(p->addr, sta->macaddr, 6);
	/* Reported BSS for specific STA */
	p->bssid_nr = 1;
	p->sta_link_metrics_bssid = calloc(p->bssid_nr, sizeof(*p->sta_link_metrics_bssid));
	if (p->sta_link_metrics_bssid) {
		for (i = 0; i < p->bssid_nr; i++) {
			memcpy(p->sta_link_metrics_bssid[i].bssid, bss->bssid, 6);
			p->sta_link_metrics_bssid[i].time_delta = sta->conn_time;
			p->sta_link_metrics_bssid[i].dl_mac_data_rate = sta->dl_est_thput;
			p->sta_link_metrics_bssid[i].ul_mac_data_rate = sta->ul_est_thput;
			p->sta_link_metrics_bssid[i].ul_rcpi = sta->rssi;
		}
	}
}

static void get_sta_traffic_stats(struct wifi_network_device *own,
				struct tlv_assoc_sta_traffic_stats *p,
				uint32_t radio_index,
				uint32_t bss_index, uint32_t sta_index)
{
	struct wifi_radio_element *radio = own->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;
	struct wifi_sta_element *sta = bss->stalist + sta_index;

	memcpy(p->addr, sta->macaddr, 6);
	p->bytes_sent = sta->tx_bytes;
	p->bytes_received = sta->rx_bytes;
	p->packets_sent = sta->tx_pkts;
	p->packets_received = sta->rx_pkts;
	p->tx_packets_err = sta->tx_errors;
	p->rx_packets_err = sta->rx_errors;
}

static void get_sta_ext_link_metrics(struct wifi_network_device *own,
				struct tlv_assoc_sta_ext_link_metric *p,
				uint32_t radio_index,
				uint32_t bss_index, uint32_t sta_index)
{
	int i;
	struct wifi_radio_element *radio = own->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;
	struct wifi_sta_element *sta = bss->stalist + sta_index;

	memcpy(p->mac, sta->macaddr, 6);
	/* Reported BSS for specific STA */
	p->nbr_bssid = 1;
	p->data = calloc(p->nbr_bssid, sizeof(*p->data));
	if (p->data) {
		for (i = 0; i < p->nbr_bssid; i++) {
			memcpy(p->data[i].bssid, bss->bssid, 6);
			p->data[i].last_data_dwn_rate = sta->dl_rate;
			p->data[i].last_data_up_rate = sta->ul_rate;
			p->data[i].uti_recv = sta->ul_utilization;
			p->data[i].uti_transmit = sta->dl_utilization;
		}
	}
}

static void get_oper_ch_report(struct wifi_network_device *own,
				struct tlv_oper_ch_report *p,
				uint32_t radio_index)
{
	int i;
	struct wifi_radio_element *radio = own->radiolist + radio_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	p->op_ch_op_class_nr = radio->num_curr_opclass;

	if (p->op_ch_op_class_nr > 0)
		p->op_ch_op_class = calloc(p->op_ch_op_class_nr, sizeof(*p->op_ch_op_class));

	if (!p->op_ch_op_class) {
		p->op_ch_op_class_nr = 0;
		return;
	}
	for (i = 0; i < p->op_ch_op_class_nr; i++) {
		p->op_ch_op_class[i].op_class = radio->curr_opclass[i].id;
		p->op_ch_op_class[i].channel = radio->curr_opclass[i].channel;
		p->curr_tx_power = radio->curr_opclass[i].txpower;
	}
}

static void get_backhaul_sta_radio_cap(struct wifi_network_device *own,
					struct tlv_backhaul_sta_radio_cap *p,
					uint32_t radio_index)
{
	struct wifi_radio_element *radio = own->radiolist + radio_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	p->mac_include = 1;
	if (p->mac_include)
		memcpy(p->mac, radio->bksta.macaddr, 6);
}

static void get_unassoc_sta_link_metrics(struct wifi_network_device *own,
				struct tlv_unassoc_sta_link_metrics_resp *p)
{
	/* FIXME: Currently set to the 0/NULL from 17.2.6 */
}

static void get_client_info(struct tlv_client_info *p,
				uint8_t *bss, uint8_t *client)
{
	memcpy(p->bssid, bss, 6);
	memcpy(p->client_addr, client, 6);
}

static void get_client_cap_report(struct tlv_client_cap_report *p,
					struct wifi_assoc_events *assoc_ev)
{
	p->result_code = 0x00;

	// TODO: Need to map with 9-23/9-31
	// 1 byte: length
	// 1 byte: ht; 6 byte: vht
	p->frame_body = (uint8_t *)calloc(8, sizeof(uint8_t));
	if (p->frame_body) {
		*(p->frame_body) = 1 + 1 + 6;
		memcpy(p->frame_body + 1, &assoc_ev->caps.ht, 1);
		memcpy(p->frame_body + 2, assoc_ev->caps.vht, 6);
	}
}

static void get_scan_res(struct wifi_network_device *own,
			struct tlv_ch_scan_res *p, uint32_t radio_index,
			uint32_t opclass_index, uint32_t channel_index)
{
	int i;
	char bw_str[16] = {0};

	struct wifi_radio_element *radio = own->radiolist + radio_index;
	struct wifi_scanres_opclass_element *op = radio->scanlist[0].opclass_scanlist + opclass_index;
	struct wifi_scanres_channel_element *ch = op->channel_scanlist + channel_index;

	memcpy(p->radio_id, radio->macaddr, 6);
	p->op_class = op->opclass;
	p->ch = ch->channel;
	p->scan_res = CH_SCAN_STATUS_SUCCESS;

	if (p->scan_res == CH_SCAN_STATUS_SUCCESS) {
		p->time_len = 32;
		p->timestamp = calloc(p->time_len, sizeof(char));
		if (!p->timestamp) {
			p->time_len = 0;
		} else
			strncpy(p->timestamp, ch->tsp, p->time_len - 1);

		p->utilization = ch->utilization;
		p->noise = ch->anpi;
		p->nbr_neighbors = ch->num_neighbors;

		if (p->nbr_neighbors > 0)
			p->neighbor_data = calloc(p->nbr_neighbors, sizeof(*(p->neighbor_data)));

		if (p->neighbor_data) {
			for (i = 0; i < p->nbr_neighbors; i++) {
				struct wifi_scanres_neighbor_element *nbr = ch->nbrlist + i;
				memcpy(p->neighbor_data[i].bssid, nbr->bssid, 6);
				p->neighbor_data[i].ssid = strdup(nbr->ssid);
				p->neighbor_data[i].ssid_len = strlen(p->neighbor_data[i].ssid);
				p->neighbor_data[i].signal_strength = nbr->rssi;
				p->neighbor_data[i].bss_load_elm_present = 1;

				snprintf(bw_str, 15, "%d", nbr->bw);
				p->neighbor_data[i].ch_bandwidth = strdup(bw_str);
				p->neighbor_data[i].ch_band_len = strlen(p->neighbor_data[i].ch_bandwidth);

				if (p->neighbor_data[i].bss_load_elm_present == 1) {
					p->neighbor_data[i].ch_util = nbr->utilization;
					p->neighbor_data[i].sta_count = nbr->num_stations;
				}
			}
		}
	}
}

static struct tlv_timestamp *get_timestamp_tlv(struct delm_private *priv)
{
	struct tlv_timestamp *p;

	p = calloc(1, sizeof(*p));
	if (!p)
		return NULL;

	p->tlv_type = MAP_TLV_TIMESTAMP;
	p->time_len = 32;
	p->time = calloc(p->time_len, sizeof(char));
	if (!p->time) {
		free(p);
		return NULL;
	}

	get_timestamp(NULL, p->time);

	return p;
}

static void get_al_mac(struct delm_private *priv, struct tlv_al_mac *p)
{
	p->tlv_type = TLV_TYPE_AL_MAC_ADDRESS_TYPE;
	memcpy(p->al_mac_address, priv->ieee1905_macaddr, 6);
}

static void get_client_assoc_event(struct tlv_client_assoc_event *p, uint8_t *sta_mac,
					const char *event, uint8_t *bssid)
{
	p->tlv_type = MAP_TLV_CLIENT_ASSOCIATION_EVENT;
	memcpy(p->client_addr, sta_mac, 6);
	memcpy(p->bssid, bssid, 6);
	if (!strcmp(event, "connected"))
		p->assoc_event = 1;
	else if (!strcmp(event, "disconnected"))
		p->assoc_event = 0;
}

static void delm_send_cmdu_cb(struct ubus_request *req,
			      int type, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	uint16_t *mid;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	mid = (uint16_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "mid", &tmp))
		*mid = json_object_get_int(tmp);


	json_object_put(jobj);
}

int prepare_send_cmdu_blob(struct cmdu_cstruct *cmdu_data,
		struct blob_buf *b, bool is_response)
{
	char *tlv_data = NULL;
	uint16_t tlv_data_len = 1;
	uint16_t tlv_str_len = 0;
	int copy_index;
	char dst_addr[18] = { 0 };
	char *tlv_str;
	uint8_t *ss = NULL;
	uint16_t len;
	int i;

	blobmsg_add_u32(b, "type", cmdu_data->message_type);
	blobmsg_add_string(b, "egress", cmdu_data->intf_name);

	/* add message id for responses */
	if (is_response)
		blobmsg_add_u32(b, "mid", cmdu_data->message_id);

	hwaddr_ntoa(cmdu_data->origin, dst_addr);
	blobmsg_add_string(b, "dst_macaddr", dst_addr);

	fprintf(stdout, "|%s:%d|cmdu:%s|dst:%s|\n", __func__, __LINE__,
			map_stringify_cmdu_type(cmdu_data->message_type), dst_addr);

	if (cmdu_data->num_tlvs > 0) {
		for (i = 0; i < cmdu_data->num_tlvs; i++) {
			len = 0;
			ss = map_put_tlv_cstruct(cmdu_data->tlvs[i], &len);
			if (ss) {
				tlv_str = calloc(1, (2 * len + 1) * sizeof(char));
				if (!tlv_str) {
					if (tlv_data)
						free(tlv_data);

					free(ss);
					goto out;
				}

				btostr(ss, len, tlv_str);
				tlv_str_len = 2 * len;
				tlv_data_len += tlv_str_len;
				tlv_data = realloc(tlv_data, tlv_data_len * sizeof(char));
				if (!tlv_data) {
					fprintf(stderr, "%s:%d out of memory.!\n",
							__func__, __LINE__);
					free(ss);
					goto out;
				}

				copy_index = tlv_data_len - tlv_str_len - 1;
				memcpy(tlv_data + copy_index, tlv_str, tlv_str_len);
				free(ss);
				free(tlv_str);
			}
		}

		if (tlv_data) {
			tlv_data[tlv_data_len - 1] = '\0';
			blobmsg_add_string(b, "data", tlv_data);
			free(tlv_data);
		}
	}

	return 0;

out:

	return -1;
}

static int delm_send_cmdu_response(struct delm_private *priv,
				   struct cmdu_cstruct *cmdu_data)
{
	uint16_t msgid = 0;
	int ret = 0;
	bool is_response = true;
	struct blob_buf b = { 0 };

	fprintf(stdout, "|%s:%d| Entry \n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	ret = prepare_send_cmdu_blob(cmdu_data, &b, is_response);
	if (ret)
		goto out;


	ret = ubus_invoke(priv->ctx, priv->map1905, "send",
			  b.head, delm_send_cmdu_cb,
			  (void *)&msgid,
			  UBUS_TIMEOUT);
	if (ret) {
		fprintf(stderr, "[%s:%d] ubus call failed for |map.1905 send|",
				__func__, __LINE__);

		/* update the map.1905 object */
		priv->map1905 = delm_lookup_object(priv, "map.1905");
		if (priv->map1905 == OBJECT_INVALID) {
			fprintf(stderr, "Failed to get 'map.1905' object!\n");
		}
		goto out;
	}

out:
	blob_buf_free(&b);

	return ret;
}

static int delm_send_cmdu_request(struct delm_private *priv,
				  struct cmdu_cstruct *cmdu_data,
				  uint16_t resp_type)
{
	int ret = 0;
	bool is_response = false;
	uint16_t msgid = 0;
	struct blob_buf b = { 0 };

	fprintf(stdout, "|%s:%d| Entry \n", __func__, __LINE__);

	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	ret = prepare_send_cmdu_blob(cmdu_data, &b, is_response);
	if (ret)
		goto out;

	ret = ubus_invoke(priv->ctx, priv->map1905, "send",
			  b.head, delm_send_cmdu_cb,
			  (void *)&msgid,
			  UBUS_TIMEOUT);
	if (!ret) {
		fprintf(stderr, "%s: msg-id = %hu\n", __func__, msgid);

		/* Don't enqueue message id for topology notification,
		 * client disassociation stats & client capability qurey.
		 */
		if ((cmdu_data->message_type == CMDU_TYPE_TOPOLOGY_NOTIFICATION) ||
				(cmdu_data->message_type == CMDU_CLIENT_DISASSOCIATION_STATS) ||
				(cmdu_data->message_type == CMDU_CLIENT_CAPABILITY_QUERY) ||
				(cmdu_data->message_type == CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH))
			goto out;

		/* Store mid of the sent CMDU with an expected response type.
		 * On receiving a 'resp_type' cmdu, lookup in the ackq and
		 * drop or handle it.
		 */
		cmdu_ackq_enqueue(&priv->cmdu_ack_q, resp_type, msgid,
				  cmdu_data->origin, 2000);  //FIXME: tmo
	} else {
		fprintf(stderr, "[%s:%d] ubus call failed for |map.1905 send|",
				__func__, __LINE__);

		/* unset getter_running,
		 * if not set, refresh method won't work
		 */
		priv->getter_running = 0;

		/* update the map.1905 object */
		priv->map1905 = delm_lookup_object(priv, "map.1905");
		if (priv->map1905 == OBJECT_INVALID) {
			fprintf(stderr, "Failed to get 'map.1905' object!\n");
		}
	}

out:
	blob_buf_free(&b);

	return ret;
}

static int get_sta_index(struct wifi_bss_element *bss, uint8_t *sta)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < bss->num_stations; i++) {
		if (hwaddr_equal(bss->stalist[i].macaddr, sta)) {
			hwaddr_ntoa(sta, addr);
			fprintf(stdout, "found sta:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_opclasss_index(struct wifi_radio_element *radio, uint8_t opclass_id)
{
	int i = 0;
	struct wifi_scanres_opclass_element *opclass = NULL;

	/* TODO: Assuming single scanresult per radio */
	if (radio->scanlist) {
		for (i = 0; i < radio->scanlist[0].num_opclass_scanned; i++) {
			opclass = radio->scanlist[0].opclass_scanlist + i;
			if (opclass->opclass == opclass_id)
				return i;
		}
	}

	return -1;
}

static int get_channel_index(struct wifi_scanres_opclass_element *opclass, uint8_t channel_id)
{
	int i = 0;
	struct wifi_scanres_channel_element *ch = NULL;

	for (i = 0; i < opclass->num_channels_scanned; i++) {
		ch = opclass->channel_scanlist + i;
		if (ch->channel == channel_id)
			return i;
	}

	return -1;
}

static int get_radio_index(struct wifi_network_device *dev, uint8_t *radio_id)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < dev->num_radios; i++) {
		if (hwaddr_equal(dev->radiolist[i].macaddr, radio_id)) {
			hwaddr_ntoa(radio_id, addr);
			fprintf(stdout, "found radio:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_bss_index(struct wifi_radio_element *radio, uint8_t *bssid)
{
	int i = 0;
	char addr[18] = {0};

	for (i = 0; i < radio->num_bss; i++) {
		if (hwaddr_equal(radio->bsslist[i].bssid, bssid)) {
			hwaddr_ntoa(bssid, addr);
			fprintf(stdout, "found bss:|%s| at index %d\n", addr, i);
			return i;
		}
	}
	return -1;
}

static int get_radio_bss_sta_index(struct wifi_network_device *dev, uint8_t *sta,
		int *radio_index, int *bss_index)
{
	int i, j;
	int sta_index;

	for (i = 0; i < dev->num_radios; i++) {
		for (j = 0; j < dev->radiolist[i].num_bss; j++) {
			sta_index = get_sta_index(&dev->radiolist[i].bsslist[j], sta);
			if (sta_index != -1) {
				*radio_index = i;
				*bss_index = j;
				return sta_index;
			}
		}
	}

	return -1;
}

static int get_radio_and_bss_index(struct wifi_network_device *dev,
		uint8_t *bssid, int *radio_index)
{
	int i;
	int bss_index;

	for (i = 0; i < dev->num_radios; i++) {
		bss_index = get_bss_index(&dev->radiolist[i], bssid);
		if (bss_index != -1) {
			*radio_index = i;
			return bss_index;
		}
	}

	return -1;
}

struct wifi_network_device *delm_get_origin_dev(struct delm_private *p,
						       uint8_t *origin)
{
	int i;

	fprintf(stderr, "%s: num_devices = %d\n", __func__, p->dmi->network.num_devices);
	for (i = 0; i < p->dmi->network.num_devices; i++) {
		if (hwaddr_equal(p->dmi->network.device[i].macaddr, origin)) {
			return &p->dmi->network.device[i];
		}
	}

	return NULL;
}

static bool fill_device_info_from_tlv(struct delm_private *priv,
					struct cmdu_cstruct *cmdu_data,
					struct wifi_network_device *dev)
{
	int i, j, k;
	bool is_send = false;
	int radio_index, bss_index;
	uint8_t *tlv = NULL;

	// Set default profile to profile-1
	dev->map_profile = MAP_PROFILE_1;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_AP_OPERATIONAL_BSS:
			{
				is_send = true;
				struct tlv_ap_oper_bss *p = (struct tlv_ap_oper_bss *)tlv;
				dev->num_radios = p->radios_nr;
				if (dev->num_radios > 0) {
					dev->radiolist = (struct wifi_radio_element *)calloc(dev->num_radios,
							sizeof(struct wifi_radio_element));
				}

				if (dev->radiolist) {
					for (j = 0; j < dev->num_radios; j++) {
						memcpy(dev->radiolist[j].macaddr, p->radio[j].radio_id, 6);
						dev->radiolist[j].enabled = true;
						dev->radiolist[j].num_bss = p->radio[j].bss_nr;
						if (dev->radiolist[j].num_bss > 0) {
							dev->radiolist[j].bsslist =
								(struct wifi_bss_element *)calloc(dev->radiolist[j].num_bss,
										sizeof(struct wifi_bss_element));
						}

						if (dev->radiolist[j].bsslist) {
							for (k = 0; k < dev->radiolist[j].num_bss; k++) {
								memcpy(dev->radiolist[j].bsslist[k].bssid,
										p->radio[j].bss[k].bssid, 6);
								strncpy(dev->radiolist[j].bsslist[k].ssid,
										p->radio[j].bss[k].ssid,
										p->radio[j].bss[k].ssid_len);
								dev->radiolist[j].bsslist[k].enabled = true;
								get_timestamp(NULL, dev->radiolist[j].bsslist[k].tsp);
							}
						}
					}
				}
				break;
			}

		case MAP_TLV_ASSOCIATED_CLIENTS:
			{
				is_send = true;
				struct tlv_assoc_client *p = (struct tlv_assoc_client *)tlv;
				radio_index = 0;
				for (j = 0; j < p->bss_nr; j++) {
					bss_index = get_radio_and_bss_index(dev, p->bss[j].bssid, &radio_index);
					if (bss_index != -1) {
						dev->radiolist[radio_index].bsslist[bss_index].num_stations = p->bss[j].assoc_clients_nr;
						if (dev->radiolist[radio_index].bsslist[bss_index].num_stations > 0) {
							dev->radiolist[radio_index].bsslist[bss_index].stalist =
								(struct wifi_sta_element *)
								calloc(dev->radiolist[radio_index].bsslist[bss_index].num_stations,
										sizeof(struct wifi_sta_element));
						}

						if (dev->radiolist[radio_index].bsslist[bss_index].stalist) {
							for (k = 0; k < p->bss[j].assoc_clients_nr; k++) {
								memcpy(dev->radiolist[radio_index].bsslist[bss_index].stalist[k].macaddr,
										p->bss[j].clients[k].client_addr, 6);
								dev->radiolist[radio_index].bsslist[bss_index].stalist[k].conn_time =
									p->bss[j].clients[k].uptime;
								get_timestamp(NULL,
										dev->radiolist[radio_index].bsslist[bss_index].stalist[k].tsp);
							}
						}
					}
				}
				break;
			}

		case MAP_TLV_MULTIAP_PROFILE:
			{
				is_send = true;
				struct tlv_map_profile *p = (struct tlv_map_profile *)tlv;
				dev->map_profile = p->profile;
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;
		}
	}

	return is_send;
}

static bool fill_ap_capability(struct delm_private *priv,
				struct cmdu_cstruct *cmdu_data,
				struct wifi_network_device *dev)
{
	int i, j;
	uint8_t ht;
	uint8_t vht[6];
	bool is_send = false;
	int radio_index;
	uint8_t *tlv = NULL;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_AP_CAPABILITY:
			{
				is_send = true;
				struct tlv_ap_cap *p = (struct tlv_ap_cap *)tlv;
				/* TODO: Currently no flag described in the wifi network device/radio*/
				if (!p->non_op_ch_metric_reporting) {
					for (j = 0; j < dev->num_radios; j++) {
						dev->radiolist[j].num_unassoc_sta = 0;
						dev->radiolist[j].unassoc_stalist = NULL;
					}
				}
				break;
			}

		case MAP_TLV_AP_RADIO_BASIC_CAPABILITIES:
			{
				is_send = true;
				struct tlv_ap_radio_basic_cap *p = (struct tlv_ap_radio_basic_cap *)tlv;

				radio_index = get_radio_index(dev, p->radio_id);
				if (radio_index != -1) {
					dev->radiolist[radio_index].num_supp_opclass = p->operating_classes_nr;

					if (dev->radiolist[radio_index].num_supp_opclass > 0) {
						dev->radiolist[radio_index].supp_opclass =
							(struct wifi_opclass_supported_element *)
							calloc(dev->radiolist[radio_index].num_supp_opclass,
									sizeof(struct wifi_opclass_supported_element));
					}

					if (dev->radiolist[radio_index].supp_opclass) {
						for (j = 0; j < dev->radiolist[radio_index].num_supp_opclass; j++) {
							dev->radiolist[radio_index].supp_opclass[j].id =
								p->operating_class[j].op_class;
							dev->radiolist[radio_index].supp_opclass[j].max_txpower =
								p->operating_class[j].max_tx_power;
							dev->radiolist[radio_index].supp_opclass[j].num_exclude_channels =
								p->operating_class[j].non_op_ch_nr;

							if (p->operating_class[j].non_op_ch_nr > 0) {
								dev->radiolist[radio_index].supp_opclass[j].exclude_chanlist =
									(uint8_t *)calloc(p->operating_class[j].non_op_ch_nr, sizeof(uint8_t));

								if (dev->radiolist[radio_index].supp_opclass[j].exclude_chanlist) {
									memcpy(dev->radiolist[radio_index].supp_opclass[j].exclude_chanlist,
											p->operating_class[j].channel,
											p->operating_class[j].non_op_ch_nr);
								}
							}
						}
					}
				}
				break;
			}

		case MAP_TLV_AP_HT_CAPABILITIES:
			{
				is_send = true;
				struct tlv_ap_ht_cap *p = (struct tlv_ap_ht_cap *)tlv;
				radio_index = get_radio_index(dev, p->radio_id);
				if (radio_index != -1) {
					memset(&ht, 0, sizeof(ht));
					ht |= p->max_tx_streams_supported << 6;
					ht |= p->max_rx_streams_supported << 4;
					ht |= p->gi_20_support << 3;
					ht |= p->gi_40_support << 2;
					ht |= p->ht_40_support << 1;
					memcpy(&dev->radiolist[radio_index].caps.ht, &ht, 1);
				}
				break;
			}

		case MAP_TLV_AP_VHT_CAPABILITIES:
			{
				is_send = true;
				struct tlv_ap_vht_cap *p = (struct tlv_ap_vht_cap *)tlv;
				radio_index = get_radio_index(dev, p->radio_id);
				if (radio_index != -1) {
					memset(vht, 0, sizeof(vht));
					memcpy(&vht[0], &p->vht_tx_mcs_supported, 2);
					memcpy(&vht[2], &p->vht_rx_mcs_supported, 2);
					vht[4] |= p->max_tx_streams_supported << 5;
					vht[4] |= p->max_rx_streams_supported << 2;
					vht[4] |= p->gi_80_support << 1;
					vht[4] |= p->gi_160_support << 0;
					vht[5] |= p->vht_8080_support << 7;
					vht[5] |= p->vht_160_support << 6;
					vht[5] |= p->su_beamformer_capable << 5;
					vht[5] |= p->mu_beamformer_capable << 4;
					memcpy(dev->radiolist[radio_index].caps.vht, vht, 6);
				}
				break;
			}

		case MAP_TLV_AP_HE_CAPABILITIES:
			{
				// TODO:Currently not supported
				//struct APHECapabilitiesTLV *p = (struct APHECapabilitiesTLV *)tlv;
				break;
			}

		case MAP_TLV_METRIC_COLLECTION_INTERVAL:
			{
				is_send = true;
				struct tlv_metric_collection_interval *p = (struct tlv_metric_collection_interval *)tlv;
				dev->collect_int = p->collection_interval;
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;
		}
	}

	return is_send;
}

static bool fill_apstats_from_tlv(struct delm_private *priv,
					struct cmdu_cstruct *cmdu_data,
					struct wifi_network_device *dev)
{
	int i, j, k;
	bool is_send = false;
	uint8_t *tlv = NULL;
	int radio_index, bss_index, sta_index;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_assoc_events *assoc_ev = NULL;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_RADIO_METRICS:
			{
				is_send = true;
				struct tlv_radio_metrics *p = (struct tlv_radio_metrics *)tlv;
				radio_index = get_radio_index(dev, p->radio_id);
				if (radio_index != -1) {
					dev->radiolist[radio_index].anpi = p->noise;
					dev->radiolist[radio_index].tx_utilization = p->transmit;
					dev->radiolist[radio_index].rx_utilization = p->receive_self;
					dev->radiolist[radio_index].other_utilization = p->receive_other;
				}
				break;
			}

		case MAP_TLV_AP_METRICS:
			{
				is_send = true;
				struct tlv_ap_metrics *p = (struct tlv_ap_metrics *)tlv;
				bss_index = get_radio_and_bss_index(dev, p->bssid, &radio_index);
				if (bss_index != -1) {
					if (p->is_ac_be) {
						memcpy(dev->radiolist[radio_index].bsslist[bss_index].est_wmm_be,
								p->service_param_info_be, 3);
					}
					if (p->is_ac_bk) {
						memcpy(dev->radiolist[radio_index].bsslist[bss_index].est_wmm_bk,
								p->service_param_info_bk, 3);
					}
					if (p->is_ac_vi) {
						memcpy(dev->radiolist[radio_index].bsslist[bss_index].est_wmm_vi,
								p->service_param_info_vi, 3);
					}
					if (p->is_ac_vo) {
						memcpy(dev->radiolist[radio_index].bsslist[bss_index].est_wmm_vo,
								p->service_param_info_vo, 3);
					}
				}
				break;
			}

		case MAP_TLV_AP_EXTENDED_METRICS:
			{
				is_send = true;
				struct tlv_ap_ext_metrics *p = (struct tlv_ap_ext_metrics *)tlv;
				bss_index = get_radio_and_bss_index(dev, p->bssid, &radio_index);
				if (bss_index != -1) {
					dev->radiolist[radio_index].bsslist[bss_index].tx_ucast_bytes = p->uni_bytes_sent;
					dev->radiolist[radio_index].bsslist[bss_index].rx_ucast_bytes = p->uni_bytes_recv;
					dev->radiolist[radio_index].bsslist[bss_index].tx_mcast_bytes = p->multi_bytes_sent;
					dev->radiolist[radio_index].bsslist[bss_index].rx_mcast_bytes = p->multi_bytes_recv;
					dev->radiolist[radio_index].bsslist[bss_index].tx_bcast_bytes = p->bro_bytes_sent;
					dev->radiolist[radio_index].bsslist[bss_index].rx_bcast_bytes = p->bro_bytes_recv;
				}
				break;
			}

		case MAP_TLV_ASSOCIATED_STA_LINK_METRICS:
			{
				is_send = true;
				struct tlv_assoc_sta_link_metrics *p = (struct tlv_assoc_sta_link_metrics *)tlv;
				for (j = 0; j < p->bssid_nr; j++) {
					bss_index = get_radio_and_bss_index(dev, p->sta_link_metrics_bssid[j].bssid, &radio_index);
					if (bss_index != -1) {
						int sta_index = get_sta_index(&dev->radiolist[radio_index].bsslist[bss_index], p->addr);
						if (sta_index != -1) {
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].dl_est_thput =
								p->sta_link_metrics_bssid[j].dl_mac_data_rate;
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].ul_est_thput =
								p->sta_link_metrics_bssid[j].ul_mac_data_rate;
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].rssi =
								p->sta_link_metrics_bssid[j].ul_rcpi;
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].conn_time =
								p->sta_link_metrics_bssid[j].time_delta;

							/* Fill HT/VHT caps here using sta event list */
							for (k = 0; k < sv->num_assoc_events; k++) {
								assoc_ev = sv->assoclist + k;
								if (hwaddr_equal(assoc_ev->macaddr, p->addr) &&
										hwaddr_equal(assoc_ev->bssid, p->sta_link_metrics_bssid[j].bssid)) {
									memcpy(&dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].caps.ht,
											&assoc_ev->caps.ht, 1);
									memcpy(dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].caps.vht,
											assoc_ev->caps.vht, 6);
									break;
								}
							}
						}
					}
				}
				break;
			}

		case MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS:
			{
				is_send = true;
				struct tlv_assoc_sta_traffic_stats *p = (struct tlv_assoc_sta_traffic_stats *)tlv;
				sta_index = get_radio_bss_sta_index(dev, p->addr, &radio_index, &bss_index);
				if (sta_index != -1) {
					dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].tx_bytes = p->bytes_sent;
					dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].rx_bytes = p->bytes_received;
					dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].tx_pkts = p->packets_sent;
					dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].rx_pkts = p->packets_received;
					dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].tx_errors = p->tx_packets_err;
					dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].rx_errors = p->rx_packets_err;
				}
				break;
			}

		case MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS:
			{
				is_send = true;
				struct tlv_assoc_sta_ext_link_metric *p = (struct tlv_assoc_sta_ext_link_metric *)tlv;
				for (j = 0; j < p->nbr_bssid; j++) {
					bss_index = get_radio_and_bss_index(dev, p->data[j].bssid, &radio_index);
					if (bss_index != -1) {
						sta_index = get_sta_index(&dev->radiolist[radio_index].bsslist[bss_index], p->mac);
						if (sta_index != -1) {
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].dl_rate =
								p->data[j].last_data_dwn_rate;
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].ul_rate =
								p->data[j].last_data_up_rate;
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].dl_utilization =
								p->data[j].uti_transmit;
							dev->radiolist[radio_index].bsslist[bss_index].stalist[sta_index].ul_utilization =
								p->data[j].uti_recv;
						}
					}
				}
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;

		}
	}

	return is_send;
}

static bool fill_bksta_from_tlv(struct delm_private *priv,
				struct cmdu_cstruct *cmdu_data,
				struct wifi_network_device *dev)
{
	int i;
	bool is_send = false;
	uint8_t *tlv = NULL;
	int radio_index;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY:
			{
				is_send = true;
				struct tlv_backhaul_sta_radio_cap *p = (struct tlv_backhaul_sta_radio_cap *)tlv;
				radio_index = get_radio_index(dev, p->radio_id);
				if (p->mac_include && (radio_index != -1)) {
					memcpy(dev->radiolist[radio_index].bksta.macaddr, p->mac, 6);
				}
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in cmdu:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;
		}
	}

	return is_send;
}

static bool fill_unassocsta_from_tlv(struct delm_private *p,
					struct cmdu_cstruct *cmdu_data,
					struct wifi_network_device *dev)
{
	int i, j;
	bool is_send = false;
	uint8_t *tlv = NULL;
	int radio_index;
	int sta_count;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE:
			{
				is_send = true;
				//FIXME: take the radio index 0, Need to map
				radio_index = 0;
				struct tlv_unassoc_sta_link_metrics_resp *p = (struct tlv_unassoc_sta_link_metrics_resp *)tlv;
				dev->radiolist[radio_index].num_unassoc_sta = p->sta_nr;
				sta_count = dev->radiolist[radio_index].num_unassoc_sta;

				if (sta_count > 0) {
					dev->radiolist[radio_index].unassoc_stalist =
						(struct wifi_unassoc_sta_element *)calloc(sta_count,
								sizeof(struct wifi_unassoc_sta_element));
				}

				if (dev->radiolist[radio_index].unassoc_stalist) {
					for (j = 0; j < sta_count; j++) {
						memcpy(dev->radiolist[radio_index].unassoc_stalist[j].macaddr,
								p->unassoc_sta_link_metrics_sta[j].addr, 6);
						dev->radiolist[radio_index].unassoc_stalist[j].rssi =
							p->unassoc_sta_link_metrics_sta[j].ul_rcpi;
					}
				}
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;

		}
	}

	return is_send;
}

static bool fill_scanres_channel_from_tlv(struct delm_private *priv,
					struct cmdu_cstruct *cmdu_data,
					struct wifi_network_device *dev)
{
	int i, j;
	int radio_index;
	int opclass_index;
	int channel_index;
	bool is_send = false;
	uint8_t *tlv = NULL;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_CHANNEL_SCAN_RES:
			{
				is_send = true;
				struct tlv_ch_scan_res *p = (struct tlv_ch_scan_res *)tlv;
				radio_index = get_radio_index(dev, p->radio_id);
				if (radio_index != -1) {
					struct wifi_radio_element *radio = dev->radiolist + radio_index;
					/* TODO: Assuming single scanresult per radio */
					if (radio->num_scanresult == 0) {
						radio->num_scanresult = 1;
						radio->scanlist = (struct wifi_scanres_element *)calloc(radio->num_scanresult,
								sizeof(struct wifi_scanres_element));

						if (!radio->scanlist) {
							fprintf(stderr, "|%s:%d| out of memory\n", __func__, __LINE__);
							radio->num_scanresult = 0;
							goto error;
						}
					}
					opclass_index = get_opclasss_index(radio, p->op_class);
					if (opclass_index == -1) {
						radio->scanlist[0].num_opclass_scanned++;
						radio->scanlist[0].opclass_scanlist = realloc(radio->scanlist[0].opclass_scanlist,
								radio->scanlist[0].num_opclass_scanned *
								sizeof(struct wifi_scanres_opclass_element));

						if (!radio->scanlist[0].opclass_scanlist) {
							fprintf(stderr, "|%s:%d| out of memory\n", __func__, __LINE__);
							radio->scanlist[0].num_opclass_scanned = 0;
							goto error;
						}

						opclass_index = radio->scanlist[0].num_opclass_scanned - 1;
						memset(radio->scanlist[0].opclass_scanlist + opclass_index, 0,
								sizeof(struct wifi_scanres_opclass_element));
					}

					struct wifi_scanres_opclass_element *op = radio->scanlist[0].opclass_scanlist + opclass_index;
					op->opclass = p->op_class;
					channel_index = get_channel_index(op, p->ch);
					if (channel_index == -1) {
						op->num_channels_scanned++;
						op->channel_scanlist = realloc(op->channel_scanlist,
								op->num_channels_scanned * sizeof(struct wifi_scanres_channel_element));

						if (!op->channel_scanlist) {
							fprintf(stderr, "|%s:%d| out of memory\n", __func__, __LINE__);
							op->num_channels_scanned = 0;
							goto error;
						}

						channel_index = op->num_channels_scanned - 1;
						memset(op->channel_scanlist + channel_index, 0,
								sizeof(struct wifi_scanres_channel_element));
					}

					struct wifi_scanres_channel_element *ch = op->channel_scanlist + channel_index;
					ch->channel = p->ch;
					if (p->scan_res == CH_SCAN_STATUS_SUCCESS) {
						if (p->timestamp)
							strncpy(ch->tsp, p->timestamp, p->time_len - 1);

						ch->utilization = p->utilization;
						ch->anpi = p->noise;
						ch->num_neighbors = p->nbr_neighbors;

						if (ch->num_neighbors > 0)
							ch->nbrlist = calloc(ch->num_neighbors, sizeof(*ch->nbrlist));

						if (!ch->nbrlist) {
							ch->num_neighbors = 0;
							continue;
						}

						for (j = 0; j < ch->num_neighbors; j++) {
							struct wifi_scanres_neighbor_element *nbr = ch->nbrlist + j;
							memcpy(nbr->bssid, p->neighbor_data[j].bssid, 6);
							strncpy(nbr->ssid, p->neighbor_data[j].ssid, p->neighbor_data[j].ssid_len);
							nbr->rssi = p->neighbor_data[j].signal_strength;

							nbr->bw = atoi(p->neighbor_data[j].ch_bandwidth);

							if (p->neighbor_data[j].bss_load_elm_present == 1) {
								nbr->utilization = p->neighbor_data[j].ch_util;
								nbr->num_stations = p->neighbor_data[j].sta_count;
							}
						}
					}

				}
				break;
			}

		case MAP_TLV_TIMESTAMP:
			{
				struct tlv_timestamp *p = (struct tlv_timestamp *)tlv;
				struct wifi_radio_element *radio;

				is_send = true;
				if (p->time && (p->time_len > 0)) {
					for (j = 0; j < dev->num_radios; j++) {
						radio = dev->radiolist + j;
						if (radio->num_scanresult == 0) {
							/* TODO: Assuming single scanresult per radio */
							radio->num_scanresult = 1;
							radio->scanlist = (struct wifi_scanres_element *)calloc(radio->num_scanresult,
									sizeof(struct wifi_scanres_element));

							if (!radio->scanlist) {
								fprintf(stderr, "|%s:%d| out of memory\n", __func__, __LINE__);
								radio->num_scanresult = 0;
								goto error;
							}

							strncpy(radio->scanlist[0].tsp, p->time, p->time_len - 1);
						}
					}
				}
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;

		}
	}

error:

	return is_send;
}

static bool fill_curr_opclass_from_tlv(struct delm_private *priv,
					struct cmdu_cstruct *cmdu_data,
					struct wifi_network_device *dev)
{
	int i, j;
	bool is_send = false;
	int radio_index;
	uint8_t *tlv = NULL;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_OPERATING_CHANNEL_REPORT:
			{
				is_send = true;
				struct tlv_oper_ch_report *p = (struct tlv_oper_ch_report *)tlv;
				radio_index = get_radio_index(dev, p->radio_id);
				if (radio_index != -1) {
					dev->radiolist[radio_index].num_curr_opclass = p->op_ch_op_class_nr;

					if (dev->radiolist[radio_index].num_curr_opclass > 0) {
						dev->radiolist[radio_index].curr_opclass =
							(struct wifi_opclass_current_element *)
							calloc(dev->radiolist[radio_index].num_curr_opclass,
									sizeof(struct wifi_opclass_current_element));
					}

					if (dev->radiolist[radio_index].curr_opclass) {
						for (j = 0; j < dev->radiolist[radio_index].num_curr_opclass; j++) {
							get_timestamp(NULL,
									dev->radiolist[radio_index].curr_opclass[j].tsp);
							dev->radiolist[radio_index].curr_opclass[j].id =
								p->op_ch_op_class[j].op_class;
							dev->radiolist[radio_index].curr_opclass[j].channel =
								p->op_ch_op_class[j].channel;
							dev->radiolist[radio_index].curr_opclass[j].txpower =
								p->curr_tx_power;
						}
					}
				}
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;

		}
	}

	return is_send;
}

static int prepare_client_cap_query(struct delm_private *priv, uint8_t *sta_mac,
					uint8_t *bssid, struct cmdu_cstruct *rec_cmdu)
{
	int tlv_index = 0;
	uint16_t resp_expected = 0xffff;
	struct cmdu_cstruct *cmdu_data = NULL;
	int ret;

	cmdu_data = calloc(1, sizeof(*cmdu_data));
	if (!cmdu_data) {
		fprintf(stderr, "failed to malloc cmdu\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_CLIENT_CAPABILITY_QUERY;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);
	resp_expected = CMDU_CLIENT_CAPABILITY_REPORT;

	struct tlv_client_info *p = calloc(1, sizeof(*p));
	if (!p)
		goto error;

	p->tlv_type = MAP_TLV_CLIENT_INFO;
	memcpy(p->bssid, bssid, 6);
	memcpy(p->client_addr, sta_mac, 6);

	cmdu_data->num_tlvs = 1;
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
	if (!cmdu_data->tlvs) {
		free(p);
		cmdu_data->num_tlvs = 0;
		goto error;
	}

	cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
	ret = delm_send_cmdu_request(priv, cmdu_data, resp_expected);

error:
	map_free_cmdu(cmdu_data);
	return ret;
}

static int fill_client_capability(struct delm_private *priv,
					struct cmdu_cstruct *cmdu_data)
{
	int i, j;
	uint8_t *tlv = NULL;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_assoc_events *assoc_ev = NULL;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_CLIENT_INFO:
			{
				struct tlv_client_info *p = (struct tlv_client_info *)tlv;
				for (j = (sv->num_assoc_events - 1); j >= 0; j--) {
					if ((hwaddr_equal(sv->assoclist[j].macaddr, p->client_addr)) &&
							(hwaddr_equal(sv->assoclist[j].bssid, p->bssid))) {
						fprintf(stderr, "Found sta: " MACFMT " at index: %d\n",
								MAC2STR(p->client_addr), j);
						assoc_ev = sv->assoclist + j;
						break;
					}
				}
				break;
			}

		case MAP_TLV_CLIENT_CAPABILITY_REPORT:
			{
				struct tlv_client_cap_report *p = (struct tlv_client_cap_report *)tlv;
				if (assoc_ev && (p->result_code == 0x00)) {
					memcpy(&assoc_ev->caps.ht, p->frame_body + 1, 1);
					memcpy(assoc_ev->caps.vht, p->frame_body + 2, 6);
				}
				break;
			}

		default:
			fprintf(stdout, "unknown TLV in CMDU:|%s|", map_stringify_cmdu_type(cmdu_data->message_type));
			break;
		}
	}

	return 0;
}

static int delm_handle_topology_notification(struct delm_private *priv,
						struct cmdu_cstruct *cmdu_data)
{
	int i;
	uint8_t *tlv = NULL;
	int index;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_assoc_events *assoc_ev = NULL;
	struct wifi_disassoc_events *disassoc_ev = NULL;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
			/* no information required from this tlv */
			break;

		case MAP_TLV_CLIENT_ASSOCIATION_EVENT:
			{
				struct tlv_client_assoc_event *p = (struct tlv_client_assoc_event *)tlv;
				if (p->assoc_event == 1) {
					sv->num_assoc_events++;
					sv->assoclist = (struct wifi_assoc_events *)realloc(sv->assoclist,
							sv->num_assoc_events * sizeof(struct wifi_assoc_events));
					if (!sv->assoclist)
						return -1;

					index = sv->num_assoc_events - 1;
					assoc_ev = sv->assoclist + index;
					memset(assoc_ev, 0, sizeof(struct wifi_assoc_events));
					get_timestamp(NULL, assoc_ev->tsp);
					memcpy(assoc_ev->macaddr, p->client_addr, 6);
					memcpy(assoc_ev->bssid, p->bssid, 6);

					/* Send client capability query */
					prepare_client_cap_query(priv, p->client_addr, p->bssid, cmdu_data);

				} else if (p->assoc_event == 0) {
					sv->num_disassoc_events++;
					sv->disassoclist = (struct wifi_disassoc_events *)realloc(sv->disassoclist,
								sv->num_disassoc_events * sizeof(struct wifi_disassoc_events));
					if (!sv->disassoclist)
						return -1;

					index = sv->num_disassoc_events - 1;
					disassoc_ev = sv->disassoclist + index;
					memset(disassoc_ev, 0, sizeof(struct wifi_disassoc_events));
					get_timestamp(NULL, disassoc_ev->tsp);
					memcpy(disassoc_ev->sta_data.macaddr, p->client_addr, 6);
					memcpy(disassoc_ev->bssid, p->bssid, 6);
				}
				break;
			}

		}
	}

	return 0;
}

static int fill_disassoc_stats(struct delm_private *priv,
				struct cmdu_cstruct *cmdu_data)
{
	int i, j;
	uint8_t *tlv = NULL;
	struct wifi_disassoc_events *disassoc_ev = NULL;
	struct wifi_sta_events *sv = priv->sv;

	for (i = 0; i < cmdu_data->num_tlvs; i++) {
		tlv = cmdu_data->tlvs[i];
		switch (*tlv) {
		case MAP_TLV_STA_MAC_ADDRESS:
			{
				struct tlv_sta_mac *p = (struct tlv_sta_mac *)tlv;
				for (j = (sv->num_disassoc_events - 1); j >= 0; j--) {
					if (hwaddr_equal(sv->disassoclist[j].sta_data.macaddr, p->addr)) {
						fprintf(stderr, "Found sta: " MACFMT " at index: %d\n",
								MAC2STR(p->addr), j);
						disassoc_ev = sv->disassoclist + j;
						break;
					}
				}
				break;
			}

		case MAP_TLV_REASON_CODE:
			{
				struct tlv_reason_code *p = (struct tlv_reason_code *)tlv;
				if (disassoc_ev)
					disassoc_ev->reason_code = p->reason_code;
				break;
			}

		case MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS:
			{
				struct tlv_assoc_sta_traffic_stats *p = (struct tlv_assoc_sta_traffic_stats *)tlv;
				if (disassoc_ev) {
					disassoc_ev->sta_data.tx_bytes = p->bytes_sent;
					disassoc_ev->sta_data.rx_bytes = p->bytes_received;
					disassoc_ev->sta_data.tx_pkts = p->packets_sent;
					disassoc_ev->sta_data.rx_pkts = p->packets_received;
					disassoc_ev->sta_data.tx_errors = p->tx_packets_err;
					disassoc_ev->sta_data.rx_errors = p->rx_packets_err;
					disassoc_ev->sta_data.rtx_pkts = p->retransmission_cnt;
				}
				break;
			}
		}
	}

	return 0;
}

static void get_ieee1905_iface_neighbors_cb(struct ubus_request *req, int type,
							struct blob_attr *msg)
{
	char *str = NULL;
	char *addr = NULL;
	struct json_object *root = NULL;
	struct json_object *tmp, *nbr_array;
	uint32_t i, neighbors_nr;
	uint8_t mac[6] = {0};
	struct neighbor_query {
		bool is_present;
		uint8_t mac[6];
	} *ptr = NULL;


	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message received is NULL\n", __func__);
		return;
	}

	ptr = (struct neighbor_query *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		root = json_tokener_parse(str);
		free(str);
	}

	if (root == NULL)
		return;

	if (json_object_object_get_ex(root, "neighbors_nr", &tmp)) {
		neighbors_nr = json_object_get_int(tmp);
		if (json_object_object_get_ex(root, "neighbors", &nbr_array)) {
			for (i = 0; i < neighbors_nr; i++) {
				tmp = json_object_array_get_idx(nbr_array, i);
				addr = strdup(json_object_get_string(tmp));
				hwaddr_aton(addr, mac);
				if (hwaddr_equal(mac, ptr->mac)) {
					ptr->is_present = true;
					free(addr);
					break;
				}
				free(addr);
			}
		}
	}
	json_object_put(root);
}

static bool get_ieee1905_iface_neighbors(struct delm_private *p, char *intf,
							uint8_t *nbr_macaddr)
{
	char ieee_intf[32] = {0};
	struct neighbor_query {
		bool is_present;
		uint8_t mac[6];
	} tmp = { 0 };
	bool ret = false;
	uint32_t id;
	int res;
	struct ubus_context *ctx = NULL;

	tmp.is_present = false;
	memcpy(tmp.mac, nbr_macaddr, 6);

	snprintf(ieee_intf, 32, "ieee1905.al.%s", intf);

	ctx = ubus_connect(NULL);
	if (!ctx)
		return ret;

	if (ubus_lookup_id(ctx, ieee_intf, &id)) {
		fprintf(stderr, "[%s:%d] not present", __func__, __LINE__);
		ubus_free(ctx);
		return ret;
	}

	res = ubus_invoke(ctx, id, "neighbors", NULL,
			get_ieee1905_iface_neighbors_cb,
			(void *)&tmp, UBUS_TIMEOUT);
	if (res) {
		fprintf(stderr,
			"[%s:%d] ubus call |%s neighbors| failed\n",
			__func__, __LINE__, ieee_intf);
	}

	ret = tmp.is_present;
	ubus_free(ctx);

	return ret;
}

/* Check the ieee1905 neighbor mac in ieee1905.al.* & return the interface name */
static void get_ieee1905_intf_name(struct delm_private *p, char *intf,
							uint8_t *nbr_macaddr)
{
	int i;

	for (i = 0; i < p->iifs.num; i++) {
		if (get_ieee1905_iface_neighbors(p, p->iifs.ifname[i], nbr_macaddr)) {
			strncpy(intf, p->iifs.ifname[i], 15);
			fprintf(stderr,
				"Found neighbor in ieee1905 '%s'\n", intf);
			break;
		}
	}
}

void delm_update_datamodel(struct delm_private *p)
{
	struct wifi_data_element *dm_old;

	//TODO: notify wifi.dataelements avaiable event
	fprintf(stderr, "%s: Updating DM in Collector...\n", __func__);
	dm_old = p->dm;
	p->dm = p->dmi;
	p->dm_available = 1;

	delm_free_dm(dm_old);
	p->dmi = NULL;
	/* update controller id */
	memcpy(p->dm->network.cntlr_id, p->cntlr_id, 6);

	p->getter_running = 0;
}

int prepare_1905_query(struct delm_private *priv, uint16_t cmdu_type,
			struct wifi_network_device *dev)
{
	uint16_t resp_expected = 0xffff;
	struct cmdu_cstruct *cmdu_data;
	uint32_t total_bss = 0;
	uint32_t k = 0;
	uint32_t i, j;
	uint32_t tlv_index = 0;
	struct wifi_opclass_supported_element *opclass = NULL;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "failed to malloc cmdu\n");
		return -1;
	}

	cmdu_data->message_type = cmdu_type;
	fprintf(stderr,
		"prepare 1905 Query: >>>>> type = %s >>>>>>>>>>>>>>>>>>>>>>\n",
		map_stringify_cmdu_type(cmdu_data->message_type));

	/* No need to fill interface name for self node */
	if (!hwaddr_equal(dev->macaddr, priv->ieee1905_macaddr))
		get_ieee1905_intf_name(priv, cmdu_data->intf_name, dev->macaddr);

	fprintf(stderr, "map.1905 egress interface:|%s|\n", cmdu_data->intf_name);

	/* XXX: in delm_send_cmdu_request(), "dst_macaddr" is set from cmdu->origin */
	memcpy(cmdu_data->origin, dev->macaddr, 6);

	switch (cmdu_type) {
	case CMDU_TYPE_TOPOLOGY_QUERY:
		{
			if (priv->cfg.profile == MAP_PROFILE_2) {
				struct tlv_map_profile *p = (struct tlv_map_profile *)calloc(1,
						sizeof(struct tlv_map_profile));
				if (!p)
					goto error;

				p->tlv_type = MAP_TLV_MULTIAP_PROFILE;
				p->profile = MULTIAP_PROFILE_2;

				cmdu_data->num_tlvs = 1;
				cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
				if (!cmdu_data->tlvs) {
					cmdu_data->num_tlvs = 0;
					free(p);
					goto error;
				}

				cmdu_data->tlvs[0] = (uint8_t *)p;
			} else {
				cmdu_data->num_tlvs = 0;
				cmdu_data->tlvs = NULL;
			}

			resp_expected = CMDU_TYPE_TOPOLOGY_RESPONSE;
			break;
		}

	case CMDU_AP_CAPABILITY_QUERY:
		{
			cmdu_data->num_tlvs = 0;
			cmdu_data->tlvs = NULL;
			resp_expected = CMDU_AP_CAPABILITY_REPORT;
			break;
		}

	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
		{
			cmdu_data->num_tlvs = 0;
			cmdu_data->tlvs = NULL;
			resp_expected = CMDU_BACKHAUL_STA_CAPABILITY_REPORT;
			break;
		}

	case CMDU_CHANNEL_SELECTION_REQ:
		{
			tlv_index = 0;
			cmdu_data->num_tlvs = 2 * dev->num_radios;
			cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
			if (!cmdu_data->tlvs) {
				cmdu_data->num_tlvs = 0;
				goto error;
			}

			for (i = 0; i < dev->num_radios; i++) {
				struct tlv_channel_pref *p = (struct tlv_channel_pref *)calloc(1, sizeof(struct tlv_channel_pref));
				if (p) {
					p->tlv_type = MAP_TLV_CHANNEL_PREFERENCE;
					memcpy(p->radio_id, dev->radiolist[i].macaddr, 6);
					p->ch_preference_op_class_nr = dev->radiolist[i].num_supp_opclass;
					if (p->ch_preference_op_class_nr > 0)
						p->op_class = calloc(p->ch_preference_op_class_nr, sizeof(*p->op_class));

					if (p->op_class) {
						for (j = 0; j < p->ch_preference_op_class_nr; j++) {
							opclass = dev->radiolist[i].supp_opclass + j;
							p->op_class[j].op_class = opclass->id;
							p->op_class[j].channel_nr = 0; // include all channels
							p->op_class[j].channel_list = NULL;
							p->op_class[j].preference = 0x00;
							p->op_class[j].preference_reason = 0x00;
						}
					}

					cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
				}

				struct tlv_tx_power_limit *p1 = (struct tlv_tx_power_limit *)calloc(1, sizeof(struct tlv_tx_power_limit));

				if (p1) {
					p1->tlv_type = MAP_TLV_TRANSMIT_POWER_LIMIT;
					memcpy(p1->radio_id, dev->radiolist[i].macaddr, 6);
					p1->tx_power_limit = 30; // FIXME

					cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;
				}
			}
			resp_expected = CMDU_OPERATING_CHANNEL_REPORT;
			break;
		}

	case CMDU_AP_METRICS_QUERY:
		{
			k = 0;
			/* AP Metric Query */
			cmdu_data->num_tlvs = 1;
// #ifndef _MAP_PROFILE_2
			if (dev->map_profile == MAP_PROFILE_2) {
				/* AP Radio Identifier TLV per radio */
				cmdu_data->num_tlvs += dev->num_radios;
			}
// #endif

			cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
			if (!cmdu_data->tlvs)
				goto error;

			for (i = 0; i < dev->num_radios; i++)
				total_bss += dev->radiolist[i].num_bss;

			struct tlv_ap_metric_query *p1 =
				(struct tlv_ap_metric_query *)calloc(1, sizeof(struct tlv_ap_metric_query));

			if (!p1)
				goto error;

			p1->tlv_type = MAP_TLV_AP_METRIC_QUERY;
			p1->bssid_nr = total_bss;
			p1->ap_metric_query_bssid = calloc(p1->bssid_nr,
					sizeof(*p1->ap_metric_query_bssid));
			if (!p1->ap_metric_query_bssid) {
				free(p1);
				goto error;
			}

			for (i = 0; i < dev->num_radios; i++) {
// #ifndef _MAP_PROFILE_2
				if (dev->map_profile == MAP_PROFILE_2) {
					struct tlv_ap_radio_identifier *p = (struct tlv_ap_radio_identifier *)calloc(1,
										sizeof(struct tlv_ap_radio_identifier));

					if (p) {
						p->tlv_type = MAP_TLV_AP_RADIO_IDENTIFIER;
						memcpy(p->radio_id, dev->radiolist[i].macaddr, 6);
						cmdu_data->tlvs[i] = (uint8_t *)p;
					}
				}
// #endif

				for (j = 0; j < dev->radiolist[i].num_bss; j++) {
					memcpy(p1->ap_metric_query_bssid[k].bssid, dev->radiolist[i].bsslist[j].bssid, 6);
					k++;
				}
			}

			if (dev->map_profile == MAP_PROFILE_2)
				cmdu_data->tlvs[i] = (uint8_t *)p1;
			else
				cmdu_data->tlvs[0] = (uint8_t *)p1;

			resp_expected = CMDU_AP_METRICS_RESPONSE;
			break;
		}

	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
		// TODO: Need to create the CMDU, unassoc sta list ??
		break;

	case CMDU_CHANNEL_SCAN_REQUEST:
		{
			struct tlv_ch_scan_req *p = (struct tlv_ch_scan_req *)calloc(1, sizeof(struct tlv_ch_scan_req));
			if (!p)
				goto error;

			p->tlv_type = MAP_TLV_CHANNEL_SCAN_REQ;
			p->fresh_scan = 1;
			p->nbr_radios = dev->num_radios;

			if (p->nbr_radios > 0)
				p->radio_data = calloc(p->nbr_radios, sizeof(*p->radio_data));
			if (!p->radio_data) {
				free(p);
				goto error;
			}

			for (i = 0; i < p->nbr_radios; i++) {
				memcpy(p->radio_data[i].radio_id, dev->radiolist[i].macaddr, 6);
				p->radio_data[i].nbr_op_class = dev->radiolist[i].num_supp_opclass;

				if (p->radio_data[i].nbr_op_class > 0) {
					p->radio_data[i].op_class_data = calloc(p->radio_data[i].nbr_op_class,
							sizeof(*p->radio_data[i].op_class_data));
				}

				if (p->radio_data[i].op_class_data) {
					for (j = 0; j < p->radio_data[i].nbr_op_class; j++) {
						p->radio_data[i].op_class_data[j].op_class =
							dev->radiolist[i].supp_opclass[j].id;
						p->radio_data[i].op_class_data[j].nbr_ch = 0; // Scan all channels in operating class
						if (p->radio_data[i].op_class_data[j].nbr_ch > 0) {
							p->radio_data[i].op_class_data[j].ch = (uint8_t *)calloc(sizeof(uint8_t),
									p->radio_data[i].op_class_data[j].nbr_ch);
							if (p->radio_data[i].op_class_data[j].ch) {
								// TODO: Copy the channels data
							}
						}
					}
				}
			}

			cmdu_data->num_tlvs = 1;
			cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
			if (cmdu_data->tlvs)
				cmdu_data->tlvs[0] = (uint8_t *)p;

			resp_expected = CMDU_CHANNEL_SCAN_REPORT;
			break;
		}

	default:
		fprintf(stdout, "%s:%d cmdu %d is not handled\n", __func__, __LINE__, cmdu_type);
		break;
	}

	delm_send_cmdu_request(priv, cmdu_data, resp_expected);

error:
	map_free_cmdu(cmdu_data);

	return 0;
}

int delm_handle_map_response(struct delm_private *p, struct cmdu_cstruct *cmdu)
{
	struct wifi_network_device *dev = NULL;
	bool is_send_next_query = false;
	//int done = 0;
	int ret = 0;

	ret = cmdu_ackq_dequeue(&p->cmdu_ack_q, cmdu->message_type,
				cmdu->message_id, cmdu->origin);
	if (ret) {
		fprintf(stderr, "wfadatad: drop unexpected CMDU (mid = %d)\n",
			cmdu->message_id);
		goto out;
	}

	dev = delm_get_origin_dev(p, cmdu->origin);
	if (!dev) {
		fprintf(stderr, "wfadatad: drop CMDU response from unknown src\n");
		ret = -1;
		goto out;
	}


	switch (cmdu->message_type) {
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
		//TODO: discard response in wrong state?
		is_send_next_query = fill_device_info_from_tlv(p, cmdu, dev);

		if (is_send_next_query) {
			prepare_1905_query(p, CMDU_AP_CAPABILITY_QUERY, dev);
		}
		break;

	case CMDU_AP_CAPABILITY_REPORT:
		is_send_next_query = fill_ap_capability(p, cmdu, dev);

		if (is_send_next_query)
			prepare_1905_query(p, CMDU_AP_METRICS_QUERY, dev);
		break;

	case CMDU_AP_METRICS_RESPONSE:
		is_send_next_query = fill_apstats_from_tlv(p, cmdu, dev);

		if (is_send_next_query)
			prepare_1905_query(p, CMDU_CHANNEL_SELECTION_REQ, dev);
		break;

	case CMDU_OPERATING_CHANNEL_REPORT:
		is_send_next_query = fill_curr_opclass_from_tlv(p, cmdu, dev);

		if (dev->map_profile == MAP_PROFILE_2) {
			if (is_send_next_query)
				prepare_1905_query(p, CMDU_BACKHAUL_STA_CAPABILITY_QUERY, dev);
		} else if (dev->map_profile == MAP_PROFILE_1)
			p->pending--;
		break;

	case CMDU_BACKHAUL_STA_CAPABILITY_REPORT:
		is_send_next_query = fill_bksta_from_tlv(p, cmdu, dev);

		if (is_send_next_query)
			prepare_1905_query(p, CMDU_CHANNEL_SCAN_REQUEST, dev);
		break;

	case CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE:
		fill_unassocsta_from_tlv(p, cmdu, dev);
		break;

	case CMDU_CHANNEL_SCAN_REPORT:
		fill_scanres_channel_from_tlv(p, cmdu, dev);
		p->pending--;
		break;

	default:
		fprintf(stdout, "unknown CMDU:%d", cmdu->message_type);
		break;
	}


	/* update datamodel if success */
	if (!ret && p->pending <= 0)
		delm_update_datamodel(p);

out:
	return ret;
}

int delm_handle_autoconfig_response(struct delm_private *p,
		struct cmdu_cstruct *cmdu)
{
	bool is_cntlr = false;
	int i;
	struct tlv_supp_service *supp_serv;

	supp_serv = (struct tlv_supp_service *)extract_tlv_by_type(cmdu,
			MAP_TLV_SUPPORTED_SERVICE);
	if (!supp_serv)
		return -1;

	for (i = 0; i < supp_serv->supported_services_list; i++) {
		if (supp_serv->supported_services[i].service ==
				SUPPORTED_SERVICE_MULTIAP_CONTROLLER) {
			is_cntlr = true;
			break;
		}
	}

	if (!is_cntlr) {
		fprintf(stderr, "Autoconfig response was not from controller\n");
		return -1;
	}

	fprintf(stdout, "controller id: " MACFMT "\n", MAC2STR(cmdu->origin));

	/* store controller id & al bridge */
	memcpy(p->cntlr_id, cmdu->origin, 6);
	strncpy(p->cntlr_al_bridge, cmdu->intf_name, IFNAMESIZE - 1);

	return 0;
}

int delm_handle_map_sta_event(struct delm_private *p, struct cmdu_cstruct *cmdu)
{
	int ret = 0;
	int i;
	bool is_known_src = false;

	if (!p->dm)
		return -1;

	for (i = 0; i < p->dm->network.num_devices; i++) {
		if (hwaddr_equal(p->dm->network.device[i].macaddr, cmdu->origin)) {
			is_known_src = true;
			break;
		}
	}

	if (!is_known_src)
		return -1;

	switch (cmdu->message_type) {
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
		delm_handle_topology_notification(p, cmdu);
		break;

	case CMDU_CLIENT_DISASSOCIATION_STATS:
		fill_disassoc_stats(p, cmdu);
		break;

	case CMDU_CLIENT_CAPABILITY_REPORT:
		fill_client_capability(p, cmdu);

	default:
		fprintf(stdout, "unknown CMDU:%d", cmdu->message_type);
		break;
	}

	return ret;
}

static int prepare_topology_response(struct delm_private *priv,
					struct cmdu_cstruct *rec_cmdu)
{
	uint16_t tlv_index = 0;
	struct cmdu_cstruct *cmdu_data;
	int ret, i;
	int ieee1905_num_tlvs = 0;
	uint8_t *cur_stream = NULL;
	uint8_t *tlv = NULL;
	uint16_t tlv_len;
	struct wifi_network_device *own = NULL;
	uint8_t *query = NULL;

	// Respond only when the topology query having multi-AP profile TLV
	if (!rec_cmdu->num_tlvs)
		return -1;

	query = (uint8_t *)rec_cmdu->tlvs[0];
	if (*query != MAP_TLV_MULTIAP_PROFILE)
		return -1;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;

	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_TYPE_TOPOLOGY_RESPONSE;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	ieee1905_num_tlvs = map_count_tlvs_in_stream(own->ieee1905_tlv_data);
	cmdu_data->num_tlvs = 3 + ieee1905_num_tlvs;

// #ifndef _MAP_PROFILE_2
	if (priv->cfg.profile == MAP_PROFILE_2) {
		cmdu_data->num_tlvs++; // Include Multi-AP profile TLV
	}
// #endif
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));

	if (!cmdu_data->tlvs) {
		map_free_cmdu(cmdu_data);
		return -1;
	}

	cur_stream = own->ieee1905_tlv_data;
	for (i = 0; i < ieee1905_num_tlvs; i++) {
		tlv_len = map_get_first_tlv_length(cur_stream);
		tlv = map_get_tlv_cstruct(cur_stream);

		if (tlv) {
			switch (*tlv) {
			case TLV_TYPE_DEVICE_INFORMATION_TYPE:
			case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
			case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
			case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
				cmdu_data->tlvs[tlv_index++] =  tlv;
			}
		}

		// Move to the next tlv
		cur_stream = cur_stream + tlv_len;
	}

	/* Supported Services */
	// TODO: Since no map-controller, Hardcoded to multi-ap agent
	struct tlv_supp_service *p = (struct tlv_supp_service *)calloc(1, sizeof(struct tlv_supp_service));
	if (p) {
		p->tlv_type = MAP_TLV_SUPPORTED_SERVICE;
		get_supp_services(own, p);
		cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
	}

	/* AP Operational BSS TLV */
	struct tlv_ap_oper_bss *p1 = (struct tlv_ap_oper_bss *)calloc(1, sizeof(struct tlv_ap_oper_bss));
	if (p1) {
		p1->tlv_type = MAP_TLV_AP_OPERATIONAL_BSS;
		fill_ap_operational_bss(own, p1);
		cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;
	}

	/* Associated Client TLV */
	struct tlv_assoc_client *p2 = (struct tlv_assoc_client *)calloc(1, sizeof(struct tlv_assoc_client));
	if (p2) {
		p2->tlv_type = MAP_TLV_ASSOCIATED_CLIENTS;
		get_associated_clients(own, p2);
		cmdu_data->tlvs[tlv_index++] = (uint8_t *)p2;
	}

	/* Multi-AP profile */
	// TODO: Hardcoded to profile-2
// #ifndef _MAP_PROFILE_2
	if (priv->cfg.profile == MAP_PROFILE_2) {
		struct tlv_map_profile *p3 = (struct tlv_map_profile *)calloc(1, sizeof(struct tlv_map_profile));
		if (p3) {
			p3->tlv_type = MAP_TLV_MULTIAP_PROFILE;
			get_multiap_profile(own, p3);
			cmdu_data->tlvs[tlv_index] = (uint8_t *)p3;
		}
	}
// #endif

	ret = delm_send_cmdu_response(priv, cmdu_data);
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_ap_capability_response(struct delm_private *priv,
					struct cmdu_cstruct *rec_cmdu)
{
	uint16_t tlv_index = 0;
	uint32_t j;
	struct cmdu_cstruct *cmdu_data;
	int ret;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_AP_CAPABILITY_REPORT;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	// Radio basic, HT, VHT caps per radio + AP caps
	cmdu_data->num_tlvs = 3 * own->num_radios + 1;

// #ifndef _MAP_PROFILE_2
	if (priv->cfg.profile == MAP_PROFILE_2) {
		// Channel Scan + CAC + profile-2 + Metric Collection TLV
		cmdu_data->num_tlvs += 4;
	}
// #endif
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));

	if (!cmdu_data->tlvs) {
		map_free_cmdu(cmdu_data);
		return -1;
	}

	/* AP basic Radio Capability TLV*/
	for (j = 0; j < own->num_radios; j++) {
		struct tlv_ap_radio_basic_cap *p =
			(struct tlv_ap_radio_basic_cap *)calloc(1, sizeof(struct tlv_ap_radio_basic_cap));
		if (p) {
			p->tlv_type = MAP_TLV_AP_RADIO_BASIC_CAPABILITIES;
			get_ap_radio_basic_cap(own, p, j);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
		}
	}

	/* AP capability TLV */
	struct tlv_ap_cap *p2 = (struct tlv_ap_cap *)calloc(1, sizeof(struct tlv_ap_cap));
	if (p2) {
		p2->tlv_type = MAP_TLV_AP_CAPABILITY;
		p2->op_ch_metric_reporting = 0;
		p2->non_op_ch_metric_reporting = 0;
		p2->agent_init_rcpi_steering = 1;
		cmdu_data->tlvs[tlv_index++] = (uint8_t *)p2;
	}

	/* HT Capability */
	for (j = 0; j < own->num_radios; j++) {
		struct tlv_ap_ht_cap *p3 = (struct tlv_ap_ht_cap *)calloc(1, sizeof(struct tlv_ap_ht_cap));
		if (p3) {
			p3->tlv_type = MAP_TLV_AP_HT_CAPABILITIES;
			get_ap_ht_caps(own, p3, j);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p3;
		}
	}

	/* VHT Capability */
	for (j = 0; j < own->num_radios; j++) {
		struct tlv_ap_vht_cap *p4 = (struct tlv_ap_vht_cap *)calloc(1, sizeof(struct tlv_ap_vht_cap));
		if (p4) {
			p4->tlv_type = MAP_TLV_AP_VHT_CAPABILITIES;
			get_ap_vht_caps(own, p4, j);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p4;
		}
	}

// #ifndef _MAP_PROFILE_2
	if (priv->cfg.profile == MAP_PROFILE_2) {
		/* Channel Scan Capability */
		struct tlv_ch_scan_cap *p5 = (struct tlv_ch_scan_cap *)calloc(1,
						sizeof(struct tlv_ch_scan_cap));
		if (p5) {
			p5->tlv_type = MAP_TLV_CHANNEL_SCAN_CAPABILITY;
			get_ch_scan_caps(own, p5);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p5;
		}

		/* CAC Capability */
		struct tlv_cac_cap *p6 = (struct tlv_cac_cap *)calloc(1,
						sizeof(struct tlv_cac_cap));
		if (p6) {
			p6->tlv_type = MAP_TLV_CAC_CAPABILITY;
			get_cac_caps(own, p6);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p6;
		}

		/* Profile-2 AP Capablility */
		struct tlv_profile2_ap_cap *p7 = (struct tlv_profile2_ap_cap *)calloc(1,
							sizeof(struct tlv_profile2_ap_cap));
		if (p7) {
			p7->tlv_type = MAP_TLV_PROFILE2_AP_CAPABILITY;
			get_profile2_caps(own, p7);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p7;
		}

		/* Collection Interval TLV*/
		struct tlv_metric_collection_interval *p1 = (struct tlv_metric_collection_interval *)calloc(1,
				sizeof(struct tlv_metric_collection_interval));
		if (p1) {
			p1->tlv_type = MAP_TLV_METRIC_COLLECTION_INTERVAL;
			get_collection_interval(own, p1);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;
		}
	}
// #endif

	ret = delm_send_cmdu_response(priv, cmdu_data);
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_apstats_response(struct delm_private *priv,
					struct cmdu_cstruct *rec_cmdu)
{
	uint16_t tlv_index = 0;
	uint32_t i, j, k;
	int ret;
	struct cmdu_cstruct *cmdu_data;
	struct wifi_network_device *own = NULL;
	uint8_t *query;
	uint32_t profile = priv->cfg.profile;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1,
			sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_AP_METRICS_RESPONSE;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	if (rec_cmdu->num_tlvs == 0)
		goto out;

	/* calculate number of TLVs in response
	 * based on the receied CMDU.
	 */
	for (i = 0; i < rec_cmdu->num_tlvs; i++) {
		query = (uint8_t *)rec_cmdu->tlvs[i];

		switch (*query) {
		case MAP_TLV_AP_RADIO_IDENTIFIER:
			{
				int radio_index;
				struct tlv_ap_radio_identifier *tmp =
					(struct tlv_ap_radio_identifier *)query;

				radio_index = get_radio_index(own, tmp->radio_id);
				if ((radio_index != -1) && (profile == MAP_PROFILE_2)) {
					/* Radio Metrics TLV */
					cmdu_data->num_tlvs++;
				}
				break;
			}

		case MAP_TLV_AP_METRIC_QUERY:
			{
				int radio_index, bss_index;
				struct wifi_bss_element *bss;
				struct tlv_ap_metric_query *tmp =
					(struct tlv_ap_metric_query *)query;

				for (j = 0; j < tmp->bssid_nr; j++) {
					bss_index = get_radio_and_bss_index(own,
							tmp->ap_metric_query_bssid[j].bssid,
							&radio_index);
					if (bss_index == -1)
						continue;

					bss = own->radiolist[radio_index].bsslist + bss_index;
					/* AP Metrics TLV */
					cmdu_data->num_tlvs++;

					if (profile == MAP_PROFILE_2) {
						/* AP Extended Metrics TLV */
						cmdu_data->num_tlvs++;
					}

					/* Associated STA Traffic Stats TLV,
					 * Associated STA Link Metrics TLV
					 */
					cmdu_data->num_tlvs += 2 * bss->num_stations;

					if (profile == MAP_PROFILE_2) {
						/* Associated STA Extended Link Metrics TLV */
						cmdu_data->num_tlvs += bss->num_stations;
					}
				}
				break;
			}

		default:
			fprintf(stderr, "unknown TLV in CMDU: %s\n",
					map_stringify_cmdu_type(rec_cmdu->message_type));
		}
	}

	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs,
			sizeof(uint8_t *));
	if (!cmdu_data->tlvs) {
		cmdu_data->num_tlvs = 0;
		goto out;
	}

	/* prepare response based on received CMDU */
	for (i = 0; i < rec_cmdu->num_tlvs; i++) {
		query = (uint8_t *)rec_cmdu->tlvs[i];

		switch (*query) {
		case MAP_TLV_AP_RADIO_IDENTIFIER:
			{
				int radio_index;
				struct tlv_ap_radio_identifier *tmp =
					(struct tlv_ap_radio_identifier *)query;

				radio_index = get_radio_index(own, tmp->radio_id);
				if ((radio_index != -1) && (profile == MAP_PROFILE_2)) {
					/* Radio Metrics TLV */
					struct tlv_radio_metrics *p1 = calloc(1, sizeof(*p1));
					if (!p1)
						goto out;

					p1->tlv_type = MAP_TLV_RADIO_METRICS;
					get_radio_metrics(own, p1, radio_index);
					cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;
				}

				break;
			}

		case MAP_TLV_AP_METRIC_QUERY:
			{
				int radio_index, bss_index;
				struct wifi_bss_element *bss;
				struct tlv_ap_metric_query *tmp =
					(struct tlv_ap_metric_query *)query;

				for (j = 0; j < tmp->bssid_nr; j++) {
					bss_index = get_radio_and_bss_index(own,
							tmp->ap_metric_query_bssid[j].bssid,
							&radio_index);

					if (bss_index == -1)
						continue;

					bss = own->radiolist[radio_index].bsslist + bss_index;
					/* AP Metrics TLV */
					struct tlv_ap_metrics *p2 = calloc(1, sizeof(*p2));
					if (!p2)
						goto out;

					p2->tlv_type = MAP_TLV_AP_METRICS;
					get_ap_metrics(own, p2, radio_index, bss_index);
					cmdu_data->tlvs[tlv_index++] = (uint8_t *)p2;

					/* Extended AP Metrics TLV */
					if (profile == MAP_PROFILE_2) {
						struct tlv_ap_ext_metrics *p3 = calloc(1, sizeof(*p3));
						if (!p3)
							goto out;

						p3->tlv_type = MAP_TLV_AP_EXTENDED_METRICS;
						get_ap_ext_metrics(own, p3, radio_index, bss_index);
						cmdu_data->tlvs[tlv_index++] = (uint8_t *)p3;
					}

					for (k = 0; k < bss->num_stations; k++) {
						/* Associated STA Link Metrics TLV */
						struct tlv_assoc_sta_link_metrics *p4 =
							calloc(1, sizeof(*p4));

						if (!p4)
							goto out;

						p4->tlv_type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS;
						get_sta_link_metrics(own, p4, radio_index,
								bss_index, k);
						cmdu_data->tlvs[tlv_index++] = (uint8_t *)p4;


						/* Associated STA Traffic Stats TLV*/
						struct tlv_assoc_sta_traffic_stats *p5 =
							calloc(1, sizeof(*p5));

						if (!p5)
							goto out;

						p5->tlv_type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS;
						get_sta_traffic_stats(own, p5, radio_index,
								bss_index, k);
						cmdu_data->tlvs[tlv_index++] = (uint8_t *)p5;

						if (profile == MAP_PROFILE_2) {
							/* Associated STA Extended Link Metrics TLV */
							struct tlv_assoc_sta_ext_link_metric *p6 =
								calloc(1, sizeof(*p6));
							if (!p6)
								goto out;

							p6->tlv_type = MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS;
							get_sta_ext_link_metrics(own, p6, radio_index,
									bss_index, k);
							cmdu_data->tlvs[tlv_index++] = (uint8_t *)p6;
						}
					}
				}
				break;
			}
		}
	}


	ret = delm_send_cmdu_response(priv, cmdu_data);
out:
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_backhaul_response(struct delm_private *priv,
					struct cmdu_cstruct *rec_cmdu)
{
	int ret;
	uint32_t i;
	uint16_t tlv_index = 0;
	struct cmdu_cstruct *cmdu_data = NULL;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_BACKHAUL_STA_CAPABILITY_REPORT;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	cmdu_data->num_tlvs = own->num_radios;
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));

	if (!cmdu_data->tlvs) {
		map_free_cmdu(cmdu_data);
		return -1;
	}

	for (i = 0; i < own->num_radios; i++) {
		struct tlv_backhaul_sta_radio_cap *p = (struct tlv_backhaul_sta_radio_cap *)calloc(1,
				sizeof(struct tlv_backhaul_sta_radio_cap));
		if (p) {
			p->tlv_type = MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY;
			get_backhaul_sta_radio_cap(own, p, i);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
		}
	}

	ret = delm_send_cmdu_response(priv, cmdu_data);
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_unassoc_sta_response(struct delm_private *priv,
					struct cmdu_cstruct *rec_cmdu)
{
	uint16_t tlv_index = 0;
	int ret;
	struct cmdu_cstruct *cmdu_data = NULL;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	cmdu_data->num_tlvs = 1;
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));

	if (!cmdu_data->tlvs) {
		map_free_cmdu(cmdu_data);
		return -1;
	}

	struct tlv_unassoc_sta_link_metrics_resp *p =
		(struct tlv_unassoc_sta_link_metrics_resp *)
		calloc(1, sizeof(struct tlv_unassoc_sta_link_metrics_resp));
	if (p) {
		p->tlv_type = MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE;
		get_unassoc_sta_link_metrics(own, p);
		cmdu_data->tlvs[tlv_index] = (uint8_t *)p;
	}

	ret = delm_send_cmdu_response(priv, cmdu_data);
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_operating_channel_response(struct delm_private *priv,
						struct cmdu_cstruct *rec_cmdu)
{
	uint32_t i;
	uint16_t tlv_index = 0;
	int ret;
	struct cmdu_cstruct *cmdu_data;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));

	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_OPERATING_CHANNEL_REPORT;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	cmdu_data->num_tlvs = own->num_radios;
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));

	if (!cmdu_data->tlvs) {
		map_free_cmdu(cmdu_data);
		return -1;
	}

	for (i = 0; i < own->num_radios; i++) {
		struct tlv_oper_ch_report *p =
			(struct tlv_oper_ch_report *)calloc(1, sizeof(struct tlv_oper_ch_report));
		if (p) {
			p->tlv_type = MAP_TLV_OPERATING_CHANNEL_REPORT;
			get_oper_ch_report(own, p, i);
			cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
		}
	}

	ret = delm_send_cmdu_response(priv, cmdu_data);

	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_scanres_channel_response(struct delm_private *priv,
						struct cmdu_cstruct *rec_cmdu)
{
	uint32_t i, j, k;
	int ret;
	int radio_index;
	int opclass_index;
	int channel_index;
	uint8_t *tmp = NULL;
	uint8_t **tlv_data = NULL;
	uint8_t num_tlvs = 0;
	struct cmdu_cstruct *cmdu_data;
	struct tlv_timestamp *p1;
	struct wifi_network_device *own = NULL;

	if (priv->dm_available)
		own = &priv->dm->network.device[0];

	if (!own)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_CHANNEL_SCAN_REPORT;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	tmp = rec_cmdu->tlvs[0];
	if (*tmp != MAP_TLV_CHANNEL_SCAN_REQ) {
		goto error;
	}

	struct tlv_ch_scan_req *query = (struct tlv_ch_scan_req *)tmp;

	/* timestamp TLV */
	num_tlvs = 1;
	tlv_data = (uint8_t **)calloc(num_tlvs, sizeof(uint8_t *));
	if (!tlv_data)
		goto error;

	p1 = get_timestamp_tlv(priv);
	if (!p1)
		goto error;

	tlv_data[num_tlvs - 1] = (uint8_t *)p1;

	/* Scan result TLV per radio */
	for (i = 0; i < query->nbr_radios; i++) {
		radio_index = get_radio_index(own, query->radio_data[i].radio_id);
		if (radio_index != -1) {
			struct wifi_radio_element *radio = own->radiolist + radio_index;
			for (j = 0; j < query->radio_data[i].nbr_op_class; j++) {
				opclass_index = get_opclasss_index(radio,
						query->radio_data[i].op_class_data[j].op_class);
				if (opclass_index != -1) {
					/* TODO: Assuming single scanresult per radio */
					struct wifi_scanres_opclass_element *opclass = radio->scanlist[0].opclass_scanlist +
						opclass_index;

					if (query->radio_data[i].op_class_data[j].nbr_ch > 0) {
						for (k = 0; k < query->radio_data[i].op_class_data[j].nbr_ch; k++) {
							channel_index =
								get_channel_index(opclass,
										query->radio_data[i].op_class_data[j].ch[k]);
							if (channel_index != -1) {
								struct tlv_ch_scan_res *p = (struct tlv_ch_scan_res *)calloc(1,
										sizeof(struct tlv_ch_scan_res));
								if (p) {
									p->tlv_type = MAP_TLV_CHANNEL_SCAN_RES;
									get_scan_res(own, p, radio_index, opclass_index, channel_index);
									num_tlvs++;
									tlv_data = (uint8_t **)realloc(tlv_data, sizeof(uint8_t *) * num_tlvs);
									if (tlv_data)
										tlv_data[num_tlvs - 1] = (uint8_t *)p;
								}
							}
						}
					} else {
						for (k = 0; k < opclass->num_channels_scanned; k++) {
							channel_index = k;
							struct tlv_ch_scan_res *p = (struct tlv_ch_scan_res *)calloc(1,
									sizeof(struct tlv_ch_scan_res));
							if (p) {
								p->tlv_type = MAP_TLV_CHANNEL_SCAN_RES;
								get_scan_res(own, p, radio_index, opclass_index, channel_index);
								num_tlvs++;
								tlv_data = (uint8_t **)realloc(tlv_data, sizeof(uint8_t *) * num_tlvs);
								if (tlv_data)
									tlv_data[num_tlvs - 1] = (uint8_t *)p;
							}
						}
					}
				}
			}
		}
	}

	cmdu_data->num_tlvs = num_tlvs;
	cmdu_data->tlvs = tlv_data;
	ret = delm_send_cmdu_response(priv, cmdu_data);

error:
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_client_capability_response(struct delm_private *priv,
						struct cmdu_cstruct *rec_cmdu)
{
	int i;
	struct cmdu_cstruct *cmdu_data;
	uint8_t *tmp = NULL;
	int ret;
	bool is_sta_pres = false;
	uint16_t tlv_index = 0;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_assoc_events *assoc_ev = NULL;

	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	if (rec_cmdu->num_tlvs == 0)
		goto error;

	tmp = rec_cmdu->tlvs[0];
	if (*tmp != MAP_TLV_CLIENT_INFO)
		goto error;

	struct tlv_client_info *query = (struct tlv_client_info *)tmp;

	cmdu_data->message_type = CMDU_CLIENT_CAPABILITY_REPORT;
	memcpy(cmdu_data->origin, rec_cmdu->origin, 6);
	cmdu_data->message_id = rec_cmdu->message_id;
	strcpy(cmdu_data->intf_name, rec_cmdu->intf_name);

	cmdu_data->num_tlvs = 2;

	struct tlv_client_info *p = calloc(1, sizeof(*p));
	if (!p) {
		cmdu_data->num_tlvs = 0;
		goto error;
	}

	p->tlv_type = MAP_TLV_CLIENT_INFO;
	get_client_info(p, query->bssid, query->client_addr);

	for (i = 0; i < sv->num_assoc_events; i++) {
		assoc_ev = sv->assoclist + i;
		if (hwaddr_equal(assoc_ev->macaddr, query->client_addr)
				&& hwaddr_equal(assoc_ev->bssid, query->bssid)) {
			is_sta_pres = true;
			break;
		}
	}

	struct tlv_client_cap_report *p1 = calloc(1, sizeof(*p1));
	if (!p1) {
		free(p);
		cmdu_data->num_tlvs = 0;
		goto error;
	}

	p1->tlv_type = MAP_TLV_CLIENT_CAPABILITY_REPORT;

	struct tlv_error_code *p2 = calloc(1, sizeof(*p2));
	if (!p2) {
		free(p);
		free(p1);
		cmdu_data->num_tlvs = 0;
		goto error;
	}

	p2->tlv_type = MAP_TLV_ERROR_CODE;

	if (is_sta_pres) {
		get_client_cap_report(p1, assoc_ev);
		if (p1->result_code == 0x01) {
			cmdu_data->num_tlvs++;
			p2->reason_code = 0x03;
			memcpy(p2->addr, query->client_addr, 6);
		}
	} else if (!is_sta_pres) {
		cmdu_data->num_tlvs++;
		p1->result_code = 0x01;
		p2->reason_code = 0x02;
		memcpy(p2->addr, query->client_addr, 6);
	}

	cmdu_data->tlvs = (uint8_t **)calloc(sizeof(uint8_t *),
			cmdu_data->num_tlvs);
	if (cmdu_data->tlvs) {
		cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
		cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;
		if (p1->result_code != 0x00) {
			cmdu_data->tlvs[tlv_index] = (uint8_t *)p2;
		} else {
			free(p2);
		}
	}

	ret = delm_send_cmdu_response(priv, cmdu_data);

error:
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_topology_notication(struct delm_private *priv, uint8_t *sta_mac,
					const char *event, uint8_t *bssid)
{
	struct cmdu_cstruct *cmdu_data;
	int ret;
	uint16_t tlv_index = 0;

	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_TYPE_TOPOLOGY_NOTIFICATION;
	strncpy(cmdu_data->intf_name, "br-lan", 15); //FIXME

	/* Use broadcast mac */
	hwaddr_aton("01:80:C2:00:00:13", cmdu_data->origin);

	// 1905 AL mac address tlv
	struct tlv_al_mac *p = calloc(1, sizeof(*p));
	if (!p) {
		goto error;
	}

	get_al_mac(priv, p);

	// client association event tlv
	struct tlv_client_assoc_event *p1 = calloc(1, sizeof(*p1));
	if (!p1) {
		free(p);
		goto error;
	}

	get_client_assoc_event(p1, sta_mac, event, bssid);

	cmdu_data->num_tlvs = 2;
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
	if (!cmdu_data->tlvs) {
		free(p);
		free(p1);
		cmdu_data->num_tlvs = 0;
		goto error;
	}

	cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
	cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;

	ret = delm_send_cmdu_request(priv, cmdu_data, 0x00);

error:
	map_free_cmdu(cmdu_data);

	return ret;
}

static int prepare_disassoc_stats(struct delm_private *priv,
		uint8_t *sta_mac, int reason_code)
{
	int i;
	bool is_sta = false;
	int ret;
	struct wifi_sta_events *sv = priv->sv;
	struct wifi_disassoc_events *disassoc_ev = NULL;
	uint16_t tlv_index = 0;
	struct cmdu_cstruct *cmdu_data = NULL;

	for (i = (sv->num_disassoc_events - 1); i >= 0; i--) {
		disassoc_ev = sv->disassoclist + i;
		if (hwaddr_equal(disassoc_ev->sta_data.macaddr, sta_mac)) {
			is_sta = true;
			break;
		}
	}

	if (!is_sta)
		return -1;


	cmdu_data = (struct cmdu_cstruct *)calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu_data) {
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	cmdu_data->message_type = CMDU_CLIENT_DISASSOCIATION_STATS;
	memcpy(cmdu_data->origin, priv->cntlr_id, 6);
	strncpy(cmdu_data->intf_name, priv->cntlr_al_bridge, IFNAMESIZE - 1);

	/* sta mac address tlv */
	struct tlv_sta_mac *p = calloc(1, sizeof(*p));
	if (!p)
		goto error;

	p->tlv_type = MAP_TLV_STA_MAC_ADDRESS;
	memcpy(p->addr, sta_mac, 6);

	/* reason code tlv */
	struct tlv_reason_code *p1 = calloc(1, sizeof(*p1));
	if (!p1) {
		free(p);
		goto error;
	}

	p1->tlv_type = MAP_TLV_REASON_CODE;
	p1->reason_code = reason_code;

	/* Associated sta stats tlv */
	struct tlv_assoc_sta_traffic_stats *p2 = calloc(1, sizeof(*p2));
	if (!p2) {
		free(p);
		free(p1);
		goto error;
	}

	p2->tlv_type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS;
	memcpy(p2->addr, sta_mac, 6);
	p2->bytes_sent = disassoc_ev->sta_data.tx_bytes;
	p2->bytes_received = disassoc_ev->sta_data.rx_bytes;
	p2->packets_sent = disassoc_ev->sta_data.tx_pkts;
	p2->packets_received = disassoc_ev->sta_data.rx_pkts;
	p2->tx_packets_err = disassoc_ev->sta_data.tx_errors;
	p2->rx_packets_err = disassoc_ev->sta_data.rx_errors;
	p2->retransmission_cnt = disassoc_ev->sta_data.rtx_pkts;

	cmdu_data->num_tlvs = 3;
	cmdu_data->tlvs = (uint8_t **)calloc(cmdu_data->num_tlvs, sizeof(uint8_t *));
	if (!cmdu_data->tlvs) {
		free(p);
		free(p1);
		free(p2);
		cmdu_data->num_tlvs = 0;
		goto error;
	}

	cmdu_data->tlvs[tlv_index++] = (uint8_t *)p;
	cmdu_data->tlvs[tlv_index++] = (uint8_t *)p1;
	cmdu_data->tlvs[tlv_index++] = (uint8_t *)p2;

	ret = delm_send_cmdu_request(priv, cmdu_data, 0x00);

error:
	map_free_cmdu(cmdu_data);
	return ret;
}

int prepare_autoconfig_search(struct delm_private *priv, uint8_t band)
{
	struct tlv_al_mac *p1;
	struct tlv_searched_role *p2;
	struct tlv_autoconf_freq_band *p3;
	struct tlv_supp_service *p4;
	struct tlv_searched_service *p5;
	struct tlv_map_profile *p6;
	struct cmdu_cstruct *cmdu;
	int i, tlv_index = 0;
	uint16_t resp_expected;

	cmdu = calloc(1, sizeof(*cmdu));
	if (!cmdu)
		return -1;

	cmdu->message_type = CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH;
	resp_expected = CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE;
	strncpy(cmdu->intf_name, "br-lan", IFNAMESIZE - 1);
	/* set broadcast mac */
	hwaddr_aton("01:80:c2:00:00:13", cmdu->origin);

	/*
	 * 1905.1 AL MAC address type TLV,
	 * SearchedRole TLV,
	 * AutoconfigFreqBand TLV,
	 * SupportedService TLV,
	 * SearchedService TLV.
	 */
	cmdu->num_tlvs = 1 + 1 + 1 + 1 + 1;
	if (priv->cfg.profile == MAP_PROFILE_2)
		cmdu->num_tlvs += 1; /* Multi-AP Profile TLV */

	cmdu->tlvs = (uint8_t **)calloc(cmdu->num_tlvs, sizeof(uint8_t *));
	if (!cmdu->tlvs) {
		cmdu->num_tlvs = 0;
		goto error;
	}

	/* 1905 AL mac address tlv */
	p1 = calloc(1, sizeof(*p1));
	if (!p1)
		goto error;

	p1->tlv_type = TLV_TYPE_AL_MAC_ADDRESS_TYPE;
	memcpy(p1->al_mac_address, priv->ieee1905_macaddr, 6);
	cmdu->tlvs[tlv_index++] = (uint8_t *)p1;

	/* SearchedRole TLV */
	p2 = calloc(1, sizeof(*p2));
	if (!p2)
		goto error;

	p2->tlv_type = TLV_TYPE_SEARCHED_ROLE;
	p2->role = 0x00;	/* registrar */
	cmdu->tlvs[tlv_index++] = (uint8_t *)p2;

	/* AutoconfigFreqBand TLV */
	p3 = calloc(1, sizeof(*p3));
	if (!p3)
		goto error;

	p3->tlv_type = TLV_TYPE_AUTOCONFIG_FREQ_BAND;
	p3->freq_band = band;
	cmdu->tlvs[tlv_index++] = (uint8_t *)p3;

	/* SupportedService TLV */
	p4 = calloc(1, sizeof(*p4));
	if (!p4)
		goto error;

	p4->tlv_type = MAP_TLV_SUPPORTED_SERVICE;
	p4->supported_services_list = 1;
	p4->supported_services = calloc(p4->supported_services_list,
			sizeof(*p4->supported_services));
	if (!p4->supported_services) {
		p4->supported_services_list = 0;
		map_free_tlv_cstruct((uint8_t *) p4);
		goto error;
	}

	for (i = 0; i < p4->supported_services_list; i++)
		p4->supported_services[i].service =
			SUPPORTED_SERVICE_MULTIAP_AGENT;

	cmdu->tlvs[tlv_index++] = (uint8_t *)p4;

	/* SearchedService TLV */
	p5 = calloc(1, sizeof(*p5));
	if (!p5)
		goto error;

	p5->tlv_type = MAP_TLV_SEARCHED_SERVICE;
	p5->searched_services_list = 1;
	p5->service = calloc(p5->searched_services_list, sizeof(*p5->service));
	if (!p5->service) {
		p5->searched_services_list = 0;
		map_free_tlv_cstruct((uint8_t *) p5);
		goto error;
	}
	for (i = 0; i < p5->searched_services_list; i++)
		p5->service[i] = SEARCHED_SERVICE_MULTIAP_CONTROLLER;

	cmdu->tlvs[tlv_index++] = (uint8_t *)p5;

	if (priv->cfg.profile == MAP_PROFILE_2) {
		/* Multi-AP Profile TLV */
		p6 = calloc(1, sizeof(*p6));
		if (!p6)
			goto error;

		p6->tlv_type = MAP_TLV_MULTIAP_PROFILE;
		p6->profile = MULTIAP_PROFILE_2;
		cmdu->tlvs[tlv_index++] = (uint8_t *)p6;
	}

	delm_send_cmdu_request(priv, cmdu, resp_expected);
	map_free_cmdu(cmdu);

	return 0;

error:
	map_free_cmdu(cmdu);

	return -1;
}

/* Entry point for collecting a node's wifi dataelements
 * mark the node as valid/available node
 */
int delm_collect_node(struct delm_private *p, struct wifi_network_device *dev)
{
	int ret;

	dev->is_valid_node = true;
	ret = prepare_1905_query(p, CMDU_TYPE_TOPOLOGY_QUERY, dev);

	return ret;
}

int delm_handle_map_request(struct delm_private *p, struct cmdu_cstruct *cmdu)
{
	int ret = 0;

	switch (cmdu->message_type) {
	case CMDU_TYPE_TOPOLOGY_QUERY:
		ret = prepare_topology_response(p, cmdu);
		break;
	case CMDU_AP_CAPABILITY_QUERY:
		ret = prepare_ap_capability_response(p, cmdu);
		break;
	case CMDU_AP_METRICS_QUERY:
		ret = prepare_apstats_response(p, cmdu);
		break;
	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
		ret = prepare_backhaul_response(p, cmdu);
		break;
	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
		ret = prepare_unassoc_sta_response(p, cmdu);
		break;
	case CMDU_CHANNEL_SCAN_REQUEST:
		ret = prepare_scanres_channel_response(p, cmdu);
		break;
	case CMDU_CHANNEL_SELECTION_REQ:
		ret = prepare_operating_channel_response(p, cmdu);
		break;
	case CMDU_CLIENT_CAPABILITY_QUERY:
		ret = prepare_client_capability_response(p, cmdu);
		break;
	default:
		break;
	}

	return ret;
}


int delm_handle_map_event(struct delm_private *p, struct cmdu_cstruct *cmdu)
{
	int ret = 0;


	switch (cmdu->message_type) {
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
	case CMDU_AP_CAPABILITY_REPORT:
	case CMDU_AP_METRICS_RESPONSE:
	case CMDU_BACKHAUL_STA_CAPABILITY_REPORT:
	case CMDU_UNASSOC_STA_LINK_METRIC_RESPONSE:
	case CMDU_OPERATING_CHANNEL_REPORT:
	case CMDU_CHANNEL_SCAN_REPORT:
		if (p->cfg.is_collector) {
			fprintf(stderr, "%s: priv = %p (collector = %d) cmdu = %s\n",
				__func__, p, p->cfg.is_collector,
				map_stringify_cmdu_type(cmdu->message_type));

			ret = delm_handle_map_response(p, cmdu);
		}
		break;
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		if (!p->cfg.is_sap_mode) {
			fprintf(stderr, "%s: priv = %p (%s) cmdu = %s\n",
					__func__, p,
					p->cfg.is_collector ? "collector" : "agent",
					map_stringify_cmdu_type(cmdu->message_type));

			ret = delm_handle_autoconfig_response(p, cmdu);
		}
		break;
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
	case CMDU_CLIENT_DISASSOCIATION_STATS:
	case CMDU_CLIENT_CAPABILITY_REPORT:
		if (p->cfg.is_collector) {
			fprintf(stderr, "%s: priv = %p (collector = %d) cmdu = %s\n",
					__func__, p, p->cfg.is_collector,
					map_stringify_cmdu_type(cmdu->message_type));

			ret = delm_handle_map_sta_event(p, cmdu);
		}
		break;
	case CMDU_TYPE_TOPOLOGY_QUERY:
	case CMDU_AP_CAPABILITY_QUERY:
	case CMDU_AP_METRICS_QUERY:
	case CMDU_CLIENT_CAPABILITY_QUERY:
	case CMDU_BACKHAUL_STA_CAPABILITY_QUERY:
	case CMDU_UNASSOC_STA_LINK_METRIC_QUERY:
	case CMDU_CHANNEL_SCAN_REQUEST:
	case CMDU_CHANNEL_SELECTION_REQ:
		if (!p->cfg.is_collector && !p->cfg.is_sap_mode)
			ret = delm_handle_map_request(p, cmdu);
		break;
	default:
		break;
	}

	return ret;
}

/* Handle topology change event -
 * Reset the CMDU message queue, (re)start collecting for all the nodes.
 */
int delm_handle_topology_event(struct delm_private *p, const char *type)
{
	fprintf(stdout, "%s: Topology_event:%s\n", __func__, type);

	cmdu_ackq_reset(&p->cmdu_ack_q);
	delm_free_dm(p->dmi);
	p->dmi = NULL;
	p->getter_running = 0;
	delm_collector_run(p);

	return 0;
}

static void check_ap_stalist_cb(struct ubus_request *req,
				int mtype, struct blob_attr *msg)
{
	int sta_count = 0;
	int i;
	char *str = NULL;
	char *addr = NULL;
	struct json_object *jobj = NULL;
	struct json_object *assoc_arr, *assoc_obj, *mac;
	uint8_t tmp_mac[6] = {0};
	struct sta_query {
		uint8_t sta_mac[6];
		bool is_present;
	} *tmp = NULL;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	tmp = (struct sta_query *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	/* Check the AP assoclist */
	if (json_object_object_get_ex(jobj, "assoclist", &assoc_arr)) {
		sta_count = json_object_array_length(assoc_arr);
		for (i = 0; i < sta_count; i++) {
			assoc_obj = json_object_array_get_idx(assoc_arr, i);
			if (json_object_object_get_ex(assoc_obj, "macaddr", &mac)) {
				addr = strdup(json_object_get_string(mac));
				hwaddr_aton(addr, tmp_mac);
				free(addr);
				if (hwaddr_equal(tmp_mac, tmp->sta_mac)) {
					tmp->is_present = true;
					break;
				}
			}
		}
	}

	json_object_put(jobj);
}

static void get_associated_bss_cb(struct ubus_request *req,
				int mtype, struct blob_attr *msg)
{
	char *str = NULL;
	struct json_object *jobj = NULL, *tmp;
	uint8_t *bssid;
	char *addr = NULL;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	bssid = (uint8_t *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "bssid", &tmp)) {
		addr = strdup(json_object_get_string(tmp));
		hwaddr_aton(addr, bssid);
		free(addr);
	}

	json_object_put(jobj);
}

static int delm_get_bssid(struct delm_private *p, uint8_t *mac,
				const char *ifname, uint8_t *bss_mac)
{
	uint32_t id;
	int res;
	struct blob_buf b = {0};
	char wifi_ap_obj[32] = {0};
	struct sta_query {
		uint8_t sta_mac[6];
		bool is_present;
	} tmp = { 0 };

	snprintf(wifi_ap_obj, 32, "wifi.ap.%s", ifname);

	res = ubus_lookup_id(p->ctx, wifi_ap_obj, &id);
	if (res != UBUS_STATUS_OK) {
		fprintf(stderr, "object '%s' not present", wifi_ap_obj);
		return OBJECT_INVALID;
	}

	blob_buf_init(&b, 0);
	memcpy(tmp.sta_mac, mac, 6);
	tmp.is_present = false;
	res = ubus_invoke(p->ctx, id, "assoclist", b.head,
			check_ap_stalist_cb, (void *)&tmp,
			UBUS_TIMEOUT);
	if (res) {
		fprintf(stderr, "|%s:%d| ubus call |%s assoclist| failed\n",
				__func__, __LINE__, wifi_ap_obj);
		blob_buf_free(&b);
		return -1;
	}

	if (!tmp.is_present)
		goto out;

	blob_buf_free(&b);
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	res = ubus_invoke(p->ctx, id, "status", b.head,
			get_associated_bss_cb, bss_mac,
			UBUS_TIMEOUT);

	if (res) {
		fprintf(stderr, "|%s:%d| ubus call |%s status| failed\n",
				__func__, __LINE__, wifi_ap_obj);
		blob_buf_free(&b);
		return -1;
	}

out:
	blob_buf_free(&b);
	return 0;

}

/* Handle associated/disassociated STA events */
int delm_handle_sta_event(struct delm_private *p, const char *macaddr,
			const char *ifname, const char *event)
{
	int ret = -1;
	uint8_t sta_mac[6] = {0};
	struct wifi_data_element *dm = NULL;
	int status_code = 0;
	int reason_code = 0; // FIXME
	timestamp_t ev_tsp;
	uint8_t bssid[6] = {0};
	uint8_t bss_mac[6] = {0};

	dm = p->dm;
	if (!dm)
		return -1;

	fprintf(stdout, "%s: STA '%s' is %s.\n",
			__func__, macaddr, event);

	hwaddr_aton(macaddr, sta_mac);
	get_timestamp(NULL, ev_tsp);

	if (!strcmp(event, "connected")) {
		ret = sta_eventlog_enqueue(&p->assoc_stalog, ifname, sta_mac);
		if (ret)
			return -1;

		delm_update_stalog_data(&p->assoc_stalog, ifname, sta_mac);
		delm_get_bssid(p, sta_mac, ifname, bss_mac);
		delm_get_assoc_sta(p, sta_mac, status_code, ev_tsp, bss_mac);
		ret |= delm_fill_sta_bssid(p, sta_mac, bssid);

	} else if (!strcmp(event, "disconnected")) {
		delm_get_unassoc_sta(p, sta_mac, reason_code, ev_tsp);
		ret = delm_fill_sta_bssid(p, sta_mac, bssid);
		ret |= sta_eventlog_dequeue(&p->assoc_stalog, sta_mac);
	}

	/* Don't send the notification for single-AP mode, else
	 * Multicast & loopback the 1905 topology notification;
	 * send the disassocation stats message in case of
	 * disassociation event (for profile 2 only).
	 */
	if (!ret && !p->cfg.is_sap_mode) {
		prepare_topology_notication(p, sta_mac, event, bssid);
		if ((!strcmp(event, "disconnected")) &&
				(p->cfg.profile == MAP_PROFILE_2)) {
			prepare_disassoc_stats(p, sta_mac, reason_code);
		}
	}

	return 0;
}
