/*
 * delm_getter.c
 * Implements WiFi Data Elements Agent get functions.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <net/if.h>
#include <netinet/in.h>

#include <easy/easy.h>
#include <wifi.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubus.h>

#ifndef WFA_DELM_STANDALONE
#include <map-agent/mapagent.h>
#include <map-agent/timer.h>
#else
#include "timer.h"
#endif

#include "wfa_delm.h"
#include "delm.h"
#include "debug.h"

static void agent_ubus_get_wifi_interfaces_cb(struct ubus_request *req, int type,
					      struct blob_attr *msg)
{
	struct wifi_netdev *ifs = (struct wifi_netdev *)req->priv;
	struct json_object *radio_array;
	struct json_object *json_msg;
	int i, j, k, len;
	char *json_str;

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	json_object_object_get_ex(json_msg, "radios", &radio_array);
	len = json_object_array_length(radio_array);
	if (len > WIFI_DEVICE_MAX_NUM)
		len = WIFI_DEVICE_MAX_NUM;

	for (i = 0; i < len; i++) {
		struct json_object *radio_obj, *radio_obj_name;
		struct json_object *fh_obj, *fh_obj_name;
		struct json_object *bk_obj, *bk_obj_name;
		struct json_object *fh_array;
		struct json_object *bk_array;
		int fh_len, bk_len;
		const char *radio_name;

		radio_obj = json_object_array_get_idx(radio_array, i);
		json_object_object_get_ex(radio_obj, "name", &radio_obj_name);
		radio_name = json_object_get_string(radio_obj_name);
		strncpy(ifs[i].radio, radio_name, 15);

		json_object_object_get_ex(radio_obj, "accesspoints", &fh_array);
		fh_len = json_object_array_length(fh_array);
		for (j = 0; j < fh_len; j++) {
			const char *fh_name;

			fh_obj = json_object_array_get_idx(fh_array, j);
			json_object_object_get_ex(fh_obj, "ifname", &fh_obj_name);
			fh_name = json_object_get_string(fh_obj_name);
			strncpy(ifs[i].iface[j].name, fh_name, 15);
			ifs[i].iface[j].mode = WIFI_IFACE_FH;
		}

		json_object_object_get_ex(radio_obj, "backhauls", &bk_array);
		bk_len = json_object_array_length(bk_array);
		for (k = 0; k < bk_len; k++) {
			const char *bk_name;

			bk_obj = json_object_array_get_idx(bk_array, k);
			json_object_object_get_ex(bk_obj, "ifname", &bk_obj_name);
			bk_name = json_object_get_string(bk_obj_name);
			strncpy(ifs[i].iface[k + j].name, bk_name, 15);
			ifs[i].iface[k + j].mode = WIFI_IFACE_BK;
		}
	}

out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

int agent_ubus_get_wifi_interfaces(struct ubus_context *ctx, struct wifi_netdev *ifs)
{
	struct blob_buf bb = { 0 };
	uint32_t id;
	int retry = 0;
	int retry_max = 4;
	int ret;

	do {
		ret = ubus_lookup_id(ctx, "wifi", &id);
		if (ret) {
			err("wifi object lookup failed. %s\n", ubus_strerror(ret));
			sleep(3);
		}
	} while (ret && retry++ < retry_max);

	if (ret) {
		err("Object 'wifi' not found!\n");
		return ret;
	}

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(ctx, id, "status", bb.head,
			  agent_ubus_get_wifi_interfaces_cb, ifs, 2 * 1000);

	if (ret)
		err("%s: err = %s\n", __func__, ubus_strerror(ret));

	blob_buf_free(&bb);

	return ret;
}

static int delm_get_wifi_interfaces(struct delm_private *p, struct wifi_netdev *ifs)
{
	int i, j;
	int ret;
	struct ubus_context *ctx;

	ctx = ubus_connect(NULL);
	if (!ctx) {
		fprintf(stderr, "Failed to connect to ubus\n");
		return -1;
	}

	ret = agent_ubus_get_wifi_interfaces(ctx, ifs);
	if (ret) {
		ubus_free(ctx);
		return -1;
	}

	//memcpy(ifs, wifi_ifs, sizeof(wifi_ifs));
	info("===============================================\n");
	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		char buf[256] = {0};

		j = 0;
		if (!ifs[i].radio || ifs[i].radio[0] == '\0')
			continue;

		snprintf(buf, 255, "[%d] Radio = %s  Iface = ", i, ifs[i].radio);
		while (ifs[i].iface[j].name[0] != '\0') {
			snprintf(buf + strlen(buf), 255,
				 "%s (%s)  ", ifs[i].iface[j].name,
				 ifs[i].iface[j].mode == WIFI_IFACE_BK ? "sta" : "ap");
			j++;
		}
		snprintf(buf + strlen(buf), 255, "\n");
		info(buf);
	}
	info("===============================================\n");

	ubus_free(ctx);
	return 0;
}

static int get_caps_info(char *ifname, struct wifi_radio_element *r)
{
	int ret, i;
	struct wifi_ap ap;
	struct wifi_bss *bss;
	struct wifi_caps *caps;

	memset(&ap, 0, sizeof(struct wifi_ap));
	ret = wifi_ap_info(ifname, &ap);
	bss = &ap.bss;
	caps = &ap.bss.caps;

	/* TODO: re-visit */
	if (caps->valid & WIFI_CAP_HT_VALID) {
		r->caps.ht |= ((r->tx_streams - 1) & 0x03) << 6;
		r->caps.ht |= ((r->rx_streams - 1) & 0x03) << 4;
		r->caps.ht |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_SGI20) ? 1 : 0) << 3;
		r->caps.ht |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_SGI40) ? 1 : 0) << 2;
		r->caps.ht |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_2040) ? 1 : 0) << 1;
	}

	if (caps->valid & WIFI_CAP_VHT_VALID) {
		r->caps.vht[4] |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_SGI80) ? 1 : 0) << 1;
		r->caps.vht[4] |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_SGI160) ? 1 : 0) << 0;

		r->caps.vht[5] |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_8080) ? 1 : 0) << 7;
		r->caps.vht[5] |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_160) ? 1 : 0) << 6;
		r->caps.vht[5] |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_VHT_SU_BFR) ? 1 : 0) << 5;
		r->caps.vht[5] |= (wifi_cap_isset(bss->cbitmap, WIFI_CAP_VHT_MU_BFR) ? 1 : 0) << 4;

		for (i = 0; i < 2; i++) {
			uint8_t *mcs_map = &caps->vht.supp_mcs[i*4];
			int max_mcs = -1;
			int octet = 0;
			int nss = 0;
			int l;

			for (l = 0; l < 16; l += 2) {
				uint8_t supp_mcs_mask = 0;

				if (l && !(l % 8))
					octet++;

				supp_mcs_mask = mcs_map[octet] & (0x3 << (l % 8));
				supp_mcs_mask >>= (l % 8);
				if (supp_mcs_mask == 3)
					break;

				nss++;
				if (supp_mcs_mask == 0)
					max_mcs = 7;
				else if (supp_mcs_mask == 1)
					max_mcs = 8;
				else if (supp_mcs_mask == 2)
					max_mcs = 9;
			}

			if (i == 0) {
				memcpy(&r->caps.vht[2], &max_mcs, 2);
				r->caps.vht[4] |= (nss & 0x07) << 2;
			} else {
				memcpy(&r->caps.vht[0], &max_mcs, 2);
				r->caps.vht[4] |= (nss & 0x07) << 5;
			}
		}
	}

	return ret;
}

static uint8_t rssi_to_rcpi(int rssi)
{
	if (!rssi)
		return 255;
	else if (rssi < -110)
		return 0;
	else if (rssi > 0)
		return 220;
	else
		return (rssi + 110) * 2;
}

int delm_get_sta_all(const char *ifname, struct wifi_sta_element **stalist,
						uint32_t *num_stas)
{
	struct wifi_sta_element *out;
	uint8_t stas[768] = {0};
	struct wifi_sta s;
	int nr = 128;
	int ret;
	int i;


	ret = wifi_get_assoclist(ifname, stas, &nr);
	if (ret)
		return -1;

	if (nr == 0) {
		*num_stas = 0;
		*stalist = NULL;
		return 0;
	}

	out = calloc(nr, sizeof(struct wifi_sta_element));
	if (!out)
		return -1;

	for (i = 0; i < nr; i++) {
		uint8_t *sta = &stas[i*6];
		struct wifi_sta_element *e = out + i;

		memset(&s, 0, sizeof(s));
		ret = wifi_get_sta_info(ifname, sta, &s);
		if (ret)
			continue;

		get_timestamp(NULL, e->tsp);

		memcpy(e->macaddr, sta, 6);
		e->rssi = s.rssi[0];
		e->dl_est_thput = s.rx_thput;
		e->ul_est_thput = s.tx_thput;
		e->conn_time = s.conn_time;
		e->tx_bytes = s.stats.tx_bytes;
		e->rx_bytes = s.stats.rx_bytes;
		e->tx_pkts = s.stats.tx_pkts;
		e->rx_pkts = s.stats.rx_pkts;
		e->tx_errors = s.stats.tx_fail_pkts;
		e->rtx_pkts = s.stats.tx_rtx_pkts;
		e->rx_errors = s.stats.rx_fail_pkts;

		memcpy(&e->caps.ht, s.caps.ht.byte, sizeof(e->caps.ht));
		memcpy(e->caps.vht, s.caps.vht.byte, sizeof(e->caps.vht));
		e->dl_rate = s.rx_rate.rate;
		e->ul_rate = s.tx_rate.rate;
		e->ul_utilization = s.tx_airtime;
		e->dl_utilization = s.rx_airtime;

	}

	*num_stas = nr;
	*stalist = out;

	return 0;
}

int delm_add_sta_all(struct wifi_sta_element *remote_sta, struct wifi_sta_element *sta)
{

	get_timestamp(NULL, sta->tsp);
	memcpy(sta->macaddr, remote_sta->macaddr, 6);

	sta->dl_rate = remote_sta->dl_rate;
	sta->ul_rate = remote_sta->ul_rate;
	sta->ul_utilization = remote_sta->ul_utilization;
	sta->dl_utilization = remote_sta->dl_utilization;
	sta->dl_est_thput = remote_sta->dl_est_thput;
	sta->ul_est_thput = remote_sta->ul_est_thput;
	sta->rssi = remote_sta->rssi;
	sta->conn_time = remote_sta->conn_time;
	sta->tx_bytes = remote_sta->tx_bytes;
	sta->rx_bytes = remote_sta->rx_bytes;
	sta->tx_pkts = remote_sta->tx_pkts;
	sta->rx_pkts = remote_sta->rx_pkts;
	sta->tx_errors = remote_sta->tx_errors;
	sta->rx_errors = remote_sta->rx_errors;
	sta->rtx_pkts = remote_sta->rtx_pkts;
	memcpy(&sta->caps.ht, &remote_sta->caps.ht, 1);
	memcpy(sta->caps.vht, remote_sta->caps.vht, 6);
	// TODO: ?
	// ipaddr_t ipaddr;
	// char hostname[128];
	sta->num_bcn_reports = 0;
	sta->bcn_reportlist = NULL;

	return 0;
}

int delm_get_bss(const char *ifname, struct wifi_bss_element *b)
{
	struct wifi_ap_stats ap_stats;
	ifstatus_t ifstat = 0;
	struct wifi_ap ap;
	uint8_t *bcn;
	uint8_t *ie;
	char iestr[513] = {0};
	int ret;
	int len;
	int esp_len;
	int ac; // Acess Category
	int i;

	wifi_get_ifstatus(ifname, &ifstat);

	memset(&ap, 0, sizeof(struct wifi_ap));
	ret = wifi_ap_info(ifname, &ap);
	if (ret)
		return -1;

	//wl_dump_capabilities(&bb, &bss->caps, bss->cbitmap);

	memcpy(b->bssid, ap.bss.bssid, 6);
	memcpy(b->ssid, ap.bss.ssid, 32);
	//b->num_stations = ap.bss.load.sta_count;

	/* Estimated service parameters BE/BK/VI/VO */
	// TODO: revisit (2 bytes extra allocated).
	bcn = calloc(2400 + 2, sizeof(uint8_t));
	len = 2400;
	if (bcn) {
		ret = wifi_get_beacon_ies(ifname, bcn, &len);
		if (!ret) {
			wifi_foreach_ie(ie, bcn, len) {
				btostr(ie, ie[1] + 2, iestr);
				if (!strncmp(iestr, "ff", 2)) {
					esp_len = ie[1];
					ie += 3; /* 1btye: ff, 1 byte: len, 1byte:0b */
					for (i = 1; i < esp_len; i += 3) {
						ac = (*ie) & 0x03;
						switch (ac) {
						case BK:
							b->is_ac_bk = 1;
							memcpy(b->est_wmm_bk, ie, 3);
							break;
						case BE:
							b->is_ac_be = 1;
							memcpy(b->est_wmm_be, ie, 3);
							break;
						case VI:
							b->is_ac_vi = 1;
							memcpy(b->est_wmm_vi, ie, 3);
							break;
						case VO:
							b->is_ac_vo = 1;
							memcpy(b->est_wmm_vo, ie, 3);
							break;
						}
						ie += 3;
					}
					break;
				}
				memset(iestr, 0, sizeof(iestr));
			}
		}
		free(bcn);
	}


	get_timestamp(NULL, b->tsp);
	b->enabled = !!(ifstat & IFF_UP);

	memset(&ap_stats, 0, sizeof(struct wifi_ap_stats));
	ret = wifi_ap_get_stats(ifname, &ap_stats);
	if (!ret) {
		b->tx_ucast_bytes = ap_stats.tx_ucast_pkts;	// FIXME: to bytes
		b->rx_ucast_bytes = ap_stats.rx_ucast_pkts;
		b->tx_mcast_bytes = ap_stats.tx_mcast_pkts;
		b->rx_mcast_bytes = ap_stats.rx_mcast_pkts;
		b->tx_bcast_bytes = ap_stats.tx_bcast_pkts;
		b->rx_bcast_bytes = ap_stats.rx_bcast_pkts;
	}

	ret = delm_get_sta_all(ifname, &b->stalist, &b->num_stations);

	return ret;
}

int delm_get_scanresults(const char *ifname,
			 struct wifi_scanres_element **scanlist,
			 uint32_t *num_scanresult)
{
	struct wifi_scanres_element *out;
	struct wifi_scanres_opclass_element *o;
	struct wifi_scanres_neighbor_element *nbr /*, *nbrs */;
	struct wifi_opclass supp_opclass[32] = {0};
	struct wifi_bss *bss, *bsss;
	int num_opclass;
	int num = 128;
	int idx = 0;
	int ret;
	int i, j, k;

	bsss = calloc(num, sizeof(struct wifi_bss));
	if (!bsss)
		return -1;

	ret = wifi_get_scan_results(ifname, bsss, &num);
	if (ret) {
		free(bsss);
		return -1;
	}

	ret = wifi_get_supp_opclass(ifname, &num_opclass, supp_opclass);
	if (ret) {
		free(bsss);
		return -1;
	}


	/* prepare output */
#if 0
	nbrs = calloc(num, sizeof(struct wifi_scanres_neighbor_element));
	if (!nbrs) {
		free(bsss);
		return -1;
	}
#endif

	out = calloc(1, sizeof(struct wifi_scanres_element));
	if (!out) {
		free(bsss);
		//free(nbrs);
		return -1;
	}

	get_timestamp(NULL, out->tsp);
	out->num_opclass_scanned = num_opclass;

	out->opclass_scanlist = calloc(out->num_opclass_scanned,
				sizeof(struct wifi_scanres_opclass_element));
	if (!out->opclass_scanlist) {
		out->num_opclass_scanned = 0;
		free(bsss);
		//free(nbrs);
		return -1;
	}


	for (i = 0; i < out->num_opclass_scanned; i++) {
		o = out->opclass_scanlist + i;

		if (supp_opclass[i].bw != BW20) {
			/* skip non-20Mhz opclass */
			o->opclass = supp_opclass[i].g_opclass;
			o->num_channels_scanned = 0;
			o->channel_scanlist = NULL;
			continue;
		}

		o->opclass = supp_opclass[i].g_opclass;
		o->num_channels_scanned = supp_opclass[i].opchannel.num;
		o->channel_scanlist = calloc(o->num_channels_scanned,
				sizeof(struct wifi_scanres_channel_element));
		if (!o->channel_scanlist) {
			free(bsss);
			//free(nbrs);
			o->num_channels_scanned = 0;
			return -1;
		}

		for (k = 0; k < num; k++) {
			bss = bsss + k;

			for (j =  0; j < o->num_channels_scanned; j++) {
				get_timestamp(NULL, o->channel_scanlist[j].tsp);
				o->channel_scanlist[j].channel =
						supp_opclass[i].opchannel.ch[j];
				o->channel_scanlist[j].utilization = 50;  //FIXME
				o->channel_scanlist[j].anpi = 100;        //FIXME

				if (bss->channel == supp_opclass[i].opchannel.ch[j]) {
					/* add nbr for this channel */
					struct wifi_scanres_neighbor_element *new;

					new = realloc(o->channel_scanlist[j].nbrlist,
						(o->channel_scanlist[j].num_neighbors + 1) * sizeof(*new));
					if (!new) {
						o->channel_scanlist[j].num_neighbors--;
						fprintf(stderr, "realloc scannbr failed!\n");
						break;
					}

					nbr = new + o->channel_scanlist[j].num_neighbors;
					//nbr = &nbrs[idx++];

					fprintf(stderr, "{ %s } Adding NBR-%d for CH=%d   at 0x%p\n",
							ifname, idx, bss->channel, nbr);
					idx++;
					memcpy(nbr->bssid, bss->bssid, 6);
					memcpy(nbr->ssid, bss->ssid, 32);
					nbr->rssi = rssi_to_rcpi(bss->rssi);
					nbr->bw = bss->curr_bw;
					nbr->utilization = bss->load.utilization;
					nbr->num_stations = bss->load.sta_count;

					o->channel_scanlist[j].num_neighbors++;
					o->channel_scanlist[j].nbrlist = new;

					break;
				}
			}
		}
	}

	free(bsss);
	*scanlist = out;
	*num_scanresult = 1;	//num;

	return 0;
}

int delm_get_bksta(const char *ifname, struct wifi_backhaul_element *bksta)
{
	return if_gethwaddr(ifname, bksta->macaddr);
}

int delm_get_radio(struct wifi_netdev *ifs, const char *ifname,
						struct wifi_radio_element *r)
{
	char bk_iface[WIFI_IFACE_MAX_NUM][16] = {0};
	char fh_iface[WIFI_IFACE_MAX_NUM][16] = {0};
	struct wifi_opclass supp_opclass[32] = {0};
	struct wifi_opclass opclass;
	struct wifi_radio radio;
	struct wifi_radio_stats s;
	ifstatus_t ifstat = 0;
	int num_fh_iface = 0;
	int num_bk_iface = 0;
	int num_opclass;
	int noise = 0;
	int i, j;
	int ret;

	ret = wifi_get_ifstatus(ifname, &ifstat);
	if (!ret)
		r->enabled = !!(ifstat & IFF_UP);	//FIXME: radio on/off

	ret = wifi_get_noise(ifname, &noise);
	if (!ret)
		r->anpi = rssi_to_rcpi(noise);

	memset(&radio, 0, sizeof(struct wifi_radio));
	ret = wifi_radio_info(ifname, &radio);
	//TODO: utilization, num_bss
	if (wifi_radio_get_hwaddr(ifname, r->macaddr) != 0)
		if_gethwaddr(ifname, r->macaddr);

	ret = wifi_get_country(ifname, r->country_code);
	r->tx_streams = radio.tx_streams;
	r->rx_streams = radio.rx_streams;

	/* radio stats */
	ret = wifi_radio_get_stats(ifname, &s);
	if (!ret) {
		r->rx_utilization = s.rx_bytes;
		r->tx_utilization = s.tx_bytes;
		r->total_utilization = r->rx_utilization + r->tx_utilization;
		r->other_utilization = 0;
	}

	/* current opclass */
	memset(&opclass, 0, sizeof(struct wifi_opclass));
	ret = wifi_get_curr_opclass(ifname, &opclass);
	if (!ret) {
		r->num_curr_opclass = 1;	// FIXME: from driver
		r->curr_opclass = calloc(r->num_curr_opclass,
				sizeof(struct wifi_opclass_current_element));
		if (!r->curr_opclass)
			return -1;

		get_timestamp(NULL, r->curr_opclass->tsp);
		r->curr_opclass->id = opclass.g_opclass;
		r->curr_opclass->channel = radio.channel;
		r->curr_opclass->txpower = 30;	//FIXME
	}

	/* supported opclass */
	ret = wifi_get_supp_opclass(ifname, &num_opclass, supp_opclass);
	if (!ret) {
		fprintf(stderr, "%s: Num Supported OPClass = %d\n",
						ifname, num_opclass);
		r->num_supp_opclass = num_opclass;
		r->supp_opclass = calloc(num_opclass,
				sizeof(struct wifi_opclass_supported_element));
		if (r->supp_opclass) {
			for (i = 0; i < r->num_supp_opclass; i++) {
				struct wifi_opclass_supported_element *sop =
							r->supp_opclass + i;

				sop->id = supp_opclass[i].g_opclass;
				sop->max_txpower =
					supp_opclass[i].opchannel.txpower;
				sop->num_exclude_channels = 0;	 // FIXME
				sop->exclude_chanlist = NULL;
				sop->num_channels = supp_opclass[i].opchannel.num;
				sop->chanlist = calloc(sop->num_channels, sizeof(uint8_t));
				if (sop->chanlist)
					memcpy(sop->chanlist, supp_opclass[i].opchannel.ch,
							sop->num_channels);
			}
		}
	}

	/* scan results */
	ret = delm_get_scanresults(ifname, &r->scanlist, &r->num_scanresult);

	/* bss */
	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		if (ifs[i].radio[0] != '\0' &&
			!strncmp(ifs[i].radio, ifname, strlen(ifname))) {
			break;
		}
	}

	/* return if no bss/bksta configured on this radio */
	if (i == WIFI_DEVICE_MAX_NUM)
		return 0;

	j = 0;
	while (ifs[i].iface[j].name[0] != '\0') {
		if (ifs[i].iface[j].mode == WIFI_IFACE_FH) {
			snprintf(fh_iface[num_fh_iface++], 15, "%s",
						ifs[i].iface[j].name);
		} else if (ifs[i].iface[j].mode == WIFI_IFACE_BK) {
			snprintf(bk_iface[num_bk_iface++], 15, "%s",
						ifs[i].iface[j].name);
		}
		j++;
	}

	r->num_bss = num_fh_iface;
	r->bsslist = calloc(r->num_bss, sizeof(struct wifi_bss_element));
	if (!r->bsslist)
		return -1;

	for (j = 0; j < r->num_bss; j++) {
		ret = delm_get_bss(fh_iface[j], r->bsslist + j);
	}

	/* WiFi Interface capabilities */
	/* TODO: In Multi-AP specs, this is defined per radio, while in libwifi it's per AP/BSS
	   So currently using the first BSS */
	if (num_fh_iface > 0)
		get_caps_info(fh_iface[0], r);

	/* unassociated STA */
	r->num_unassoc_sta = 0;
	r->unassoc_stalist = NULL;

	/* backhaul sta(s) */
#if 0	/* Only one 'bksta' is supported in the WiFi Data Elements Std. */
	for (j = 0; j < num_bk_iface; j++) {
		ret = delm_get_bksta(bk_iface[j], r->bkstalist + j);
	}
#else
	if (num_bk_iface > 0)
		ret = delm_get_bksta(bk_iface[0], &r->bksta);
#endif

	return 0;
}

int delm_get_device(struct wifi_netdev *ifs, struct wifi_network_device *dev)
{
	int ret = 0;
	int i;

	if (!ifs || !dev)
		return -1;

	for (i = 0; i < WIFI_DEVICE_MAX_NUM; i++) {
		if (ifs[i].radio[0] != '\0')
			dev->num_radios++;
	}

	dev->radiolist = calloc(dev->num_radios,
					sizeof(struct wifi_radio_element));
	if (!dev->radiolist)
		return -1;

	for (i = 0; i < dev->num_radios; i++) {
		ret = delm_get_radio(ifs, ifs[i].radio, dev->radiolist + i);
	}

	return ret;
}

int delm_get_assoc_sta(struct delm_private *p, uint8_t *sta_mac,
				int status_code, char *ev_tsp,
				uint8_t *bssid)
{
	int index = 0;
	struct sta_eventlog *slog = &p->assoc_stalog;
	struct sta_eventlog_entry *ev = NULL;
	struct wifi_sta_events *sv = p->sv;
	struct wifi_assoc_events *assoc_ev = NULL;

	ev = sta_eventlog_lookup(slog, sta_mac);
	if (ev == NULL) {
		fprintf(stderr, "No STA found with macaddr: " MACFMT "\n",
				MAC2STR(sta_mac));
			return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	sv->num_assoc_events++;
	index = sv->num_assoc_events - 1;
	sv->assoclist = (struct wifi_assoc_events *)realloc(sv->assoclist,
			sv->num_assoc_events * sizeof(struct wifi_assoc_events));
	assoc_ev = sv->assoclist + index;
	memset(assoc_ev, 0, sizeof(struct wifi_assoc_events));
	strncpy(assoc_ev->tsp, ev_tsp, 31);
	memcpy(ev->bssid, bssid, 6);
	memcpy(assoc_ev->macaddr, ev->sta.macaddr, 6);
	memcpy(assoc_ev->bssid, ev->bssid, 6);
	assoc_ev->status_code = status_code;
	memcpy(&assoc_ev->caps.ht, ev->sta.caps.ht.byte, sizeof(assoc_ev->caps.ht));
	memcpy(assoc_ev->caps.vht, ev->sta.caps.vht.byte, sizeof(assoc_ev->caps.vht));
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

int delm_get_unassoc_sta(struct delm_private *p, uint8_t *sta_mac,
		int reason_code, char *ev_tsp)
{
	int index = 0;
	struct sta_eventlog *slog = &p->assoc_stalog;
	struct sta_eventlog_entry *ev = NULL;
	struct wifi_sta_events *sv = p->sv;
	struct wifi_disassoc_events *disassoc_ev = NULL;

	ev = sta_eventlog_lookup(slog, sta_mac);
	if (ev == NULL) {
		fprintf(stderr, "No STA found with macaddr: " MACFMT "\n",
				MAC2STR(sta_mac));
		return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	sv->num_disassoc_events++;
	index = sv->num_disassoc_events - 1;
	sv->disassoclist = (struct wifi_disassoc_events *)realloc(sv->disassoclist,
			sv->num_disassoc_events * sizeof(struct wifi_disassoc_events));
	disassoc_ev = sv->disassoclist + index;
	memset(disassoc_ev, 0, sizeof(struct wifi_disassoc_events));
	strncpy(disassoc_ev->tsp, ev_tsp, 31);
	memcpy(disassoc_ev->bssid, ev->bssid, 6);
	memcpy(disassoc_ev->sta_data.macaddr, ev->sta.macaddr, 6);
	disassoc_ev->reason_code = reason_code;
	disassoc_ev->sta_data.tx_bytes = ev->sta.stats.tx_bytes;
	disassoc_ev->sta_data.rx_bytes = ev->sta.stats.rx_bytes;
	disassoc_ev->sta_data.tx_pkts = ev->sta.stats.tx_pkts;
	disassoc_ev->sta_data.rx_pkts = ev->sta.stats.rx_pkts;
	disassoc_ev->sta_data.tx_errors = ev->sta.stats.tx_fail_pkts;
	disassoc_ev->sta_data.rx_errors = ev->sta.stats.rx_fail_pkts;
	disassoc_ev->sta_data.rtx_pkts = ev->sta.stats.tx_retry_pkts;
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

int delm_fill_sta_bssid(struct delm_private *p, uint8_t *sta_mac, uint8_t *bssid)
{
	struct sta_eventlog *slog = &p->assoc_stalog;
	struct sta_eventlog_entry *ev = NULL;

	ev = sta_eventlog_lookup(slog, sta_mac);
	if (ev == NULL) {
		fprintf(stderr, "No STA found with macaddr: " MACFMT "\n",
				MAC2STR(sta_mac));
		return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	memcpy(bssid, ev->bssid, 6);
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

int delm_get_all(struct wifi_netdev *ifs, struct wifi_data_element *dm)
{
	return delm_get_device(ifs, &dm->network.device[0]);
}

void delm_free_dm(struct wifi_data_element *dm)
{
	int i;
	struct wifi_network *net;
	struct wifi_network_device *dev;

	if (dm) {
		fprintf(stderr, "Free old dm\n");
		net = &dm->network;
		if (net->num_devices > 0) {
			for (i = 0; i < net->num_devices; i++) {
				dev = net->device + i;
				free_dev_data(dev);
			}
		}

		free(dm);
	}
}

void *delm_getter_run(void *data)
{
	struct delm_private *p = (struct delm_private *)data;
	struct wifi_data_element *dm, *dm_old;
	struct wifi_netdev ifs[WIFI_DEVICE_MAX_NUM];
	int ret;


	p->getter_running = 1;

	memset(ifs, 0, sizeof(ifs));
	ret = delm_get_wifi_interfaces(p, ifs);
	if (ret) {
		err("%s: Failed to get wifi interfaces\n", __func__);
		p->getter_running = 0;
		return NULL;
	}

	/* This getter for this device only. So, num_devices = 1.
	 * In a Multi-AP network, each Multi-AP Agent device will run
	 * such a getter.
	 */
	dm = calloc(1, sizeof(struct wifi_data_element) +
				1 * sizeof(struct wifi_network_device));
	if (!dm) {
		p->getter_running = 0;
		return NULL;
	}

	/* dm is for this device only */
	get_timestamp(NULL, dm->network.tsp);
	dm->network.num_devices = 1;

	ret = delm_get_all(ifs, dm);
	if (!ret) {
		p->getter_running = 0;
		dm_old = p->dm;
		p->dm = dm;
		p->dm_available = 1;
		fprintf(stderr, "%s: update DM to 0x%p\n", __func__, p->dm);
		memset(p->ifs, 0, sizeof(ifs));
		memcpy(p->ifs, ifs, sizeof(ifs));

		/* fill device own al-mac & controller mac */
		memcpy(p->dm->network.device[0].macaddr,
				p->ieee1905_macaddr, 6);
		memcpy(p->dm->network.cntlr_id, p->cntlr_id, 6);

		/* now free the older dm */
		delm_free_dm(dm_old);
	}

	return NULL;
}
