/*
 * delm_ubus.c
 * WiFi Data Elements ubus methods
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pthread.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>

#include <map1905/map2.h>
#include <map1905/maputils.h>

#ifndef WFA_DELM_STANDALONE
#include <map-agent/mapagent.h>
#include <map-agent/timer.h>
#else
#include "timer.h"
#endif

#include "wfa_delm.h"
#include "delm.h"
#include "debug.h"


extern int delm_handle_map_event(struct delm_private *p,
				struct cmdu_cstruct *cmdu);
extern int delm_handle_topology_event(struct delm_private *p,
				const char *type);
extern int delm_handle_sta_event(struct delm_private *p, const char *macaddr,
				const char *ifname, const char *action);


int delm_dump(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct blob_buf bb;
	struct delm_private *p = container_of(obj, struct delm_private, obj);
	int ret;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	ret = delm_dump_all(p, &bb);
	if (ret) {
		err("DM is not ready; do a refresh!\n");
		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);

	return 0;
}

int delm_refresh(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct delm_private *p = container_of(obj, struct delm_private, obj);
	int ret = 0;

	if (!p->getter_running) {
		if (p->cfg.is_collector)
			delm_collector_run(p);
		else
			delm_getter_run(p);
	} else
		fprintf(stderr, "Getter/Collector running!!!\n");

	return ret;
}

int delm_event(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct blob_buf bb;
	struct delm_private *p = container_of(obj, struct delm_private, obj);

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	delm_event_dump(p, &bb);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

int delm_publish_object(struct delm_private *p, const char *objname)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	int ret;
	struct ubus_method m[3] = {
		UBUS_METHOD_NOARG("dump", delm_dump),
		UBUS_METHOD_NOARG("refresh", delm_refresh),
		UBUS_METHOD_NOARG("event", delm_event)
	};
	int num_methods = ARRAY_SIZE(m);

	if (!p->ctx) {
		p->ctx = ubus_connect(NULL);
		if (!p->ctx) {
			fprintf(stderr, "Failed to connect to ubus\n");
			return -1;
		}
		ubus_add_uloop(p->ctx);
	}

	obj = &p->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = objname;
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));
	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;

	ret = ubus_add_object(p->ctx, obj);
	if (ret) {
		fprintf(stderr, "Failed to add 'obj' err = %s\n",
			ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	fprintf(stderr, "Added object '%s'\n", objname);

	return 0;
}

void delm_ieee1905_cmdu_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	struct cmdu_cstruct *cmdu_data;
	uint8_t *tlv = NULL;
	char *tlvstr = NULL;
	char in_ifname[16] = {0};
	char src[18] = { 0 };
	uint8_t srcmac[6];
	int len = 0;
	uint16_t cmdu_type;
	uint16_t mid = 0;
	struct blob_attr *tb[5];
	static const struct blobmsg_policy cmdu_attrs[5] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT32 },
		[2] = { .name = "ingress", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "source", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "tlvs", .type = BLOBMSG_TYPE_STRING },
	};
	struct delm_private *priv =
			container_of(ev, struct delm_private, ieee1905_cmdu);


	fprintf(stdout, "%s: --->\n", __func__);

	blobmsg_parse(cmdu_attrs, 5, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	if (tb[0]) {
		cmdu_type = (uint16_t)blobmsg_get_u32(tb[0]);
		if (cmdu_type < 0)
			return;
	}

	if (tb[1])
		mid = (uint16_t)blobmsg_get_u32(tb[1]);


	if (tb[2]) {
		strncpy(in_ifname, blobmsg_data(tb[2]), 15);
	}

	if (tb[3]) {
		strncpy(src, blobmsg_data(tb[3]), 17);
		hwaddr_aton(src, srcmac);

	}

	if (tb[4]) {
		len = blobmsg_data_len(tb[4]);
		tlvstr = calloc(len + 1, sizeof(char));
		if (!tlvstr)
			return;

		strncpy(tlvstr, blobmsg_data(tb[4]), len);
		len = (len - 1) / 2;
		tlv = calloc(len, sizeof(uint8_t));
		if (!tlv) {
			free(tlvstr);
			return;
		}

		strtob(tlvstr, len, tlv);
	}

	cmdu_data = map_build_cmdu(cmdu_type, mid, in_ifname, srcmac, tlv);
	if (cmdu_data != NULL) {
		fprintf(stdout, "cmdu_type:%s, source:%s\n",
				map_stringify_cmdu_type(cmdu_type), src);

		delm_handle_map_event(priv, cmdu_data);
		map_free_cmdu(cmdu_data);
	} else {
		fprintf(stderr, "map_build_cmdu() failed!\n");
	}

	free(tlvstr);
	free(tlv);
}

/* This event handler for ieee1905 events(capture topology
 * response only) & get following TLVs in response,
 * (Device information TLV, Device bridging capability TLVs,
 * non-1905 neighbor device TLVs & 1905.1 neighbor device TLVs)
 */
void delm_ieee1905_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	uint8_t *tlv = NULL;
	char *tlvstr = NULL;
	uint16_t cmdu_type;
	int len = 0;
	struct blob_attr *tb[4];
	struct wifi_network_device *own = NULL;
	static const struct blobmsg_policy cmdu_attrs[5] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT32 },
		[2] = { .name = "ingress", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "tlvs", .type = BLOBMSG_TYPE_STRING },
	};
	struct delm_private *priv =
		container_of(ev, struct delm_private, ieee1905_ev);

	fprintf(stdout, "%s: --->\n", __func__);

	blobmsg_parse(cmdu_attrs, 4, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[3])
		return;

	// In case of self node, no ingress param
	if (tb[2])
		return;

	cmdu_type = (uint16_t)blobmsg_get_u32(tb[0]);
	if (cmdu_type != CMDU_TYPE_TOPOLOGY_RESPONSE)
		return;

	if (tb[3]) {
		len = blobmsg_data_len(tb[3]);
		tlvstr = calloc(len + 1, sizeof(char));
		if (!tlvstr)
			return;

		strncpy(tlvstr, blobmsg_data(tb[3]), len);
		len = (len - 1) / 2;
		tlv = calloc(len, sizeof(uint8_t));
		if (!tlv) {
			free(tlvstr);
			return;
		}

		strtob(tlvstr, len, tlv);

		if ((!priv->getter_running) && priv->dm) {
			own = &priv->dm->network.device[0];
			if (own->ieee1905_tlv_data) {
				free(own->ieee1905_tlv_data);
				own->ieee1905_tlv_data = NULL;
			}

			own->ieee1905_tlv_data = tlv;
		}

		free(tlvstr);
	}
}

void delm_topology_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	int is_1905;
	struct blob_attr *tb[4];
	static const struct blobmsg_policy neighbor_attrs[4] = {
		[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "src_macaddr", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "is1905", .type = BLOBMSG_TYPE_INT8 },
		[3] = { .name = "ingress", .type = BLOBMSG_TYPE_STRING },
	};
	struct delm_private *priv =
			container_of(ev, struct delm_private, topology_ev);


	blobmsg_parse(neighbor_attrs, 4, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[2])
		return;

	is_1905 = blobmsg_get_u8(tb[2]);
	if (!is_1905)
		return;

	delm_handle_topology_event(priv, type);
}

void delm_sta_event_handler(struct ubus_context *ctx,
				struct ubus_event_handler *ev,
				const char *type,
				struct blob_attr *msg)
{
	char ifname[16] = {0};
	char event[16] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add, del;
	struct delm_private *priv =
			container_of(ev, struct delm_private, sta_ev);

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
		struct blob_attr *data[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		};
		char macaddr[18] = {0};
		char bssid[18] = {0};

		blobmsg_parse(data_attr, 2, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(macaddr, blobmsg_data(data[0]), sizeof(macaddr) - 1);

		if (data[1])
			strncpy(bssid, blobmsg_data(data[1]), sizeof(bssid) - 1);

		delm_handle_sta_event(priv, macaddr, ifname, event);
	}
}

int delm_register_events(struct delm_private *p)
{
	int ret;

	p->ieee1905_cmdu.cb = delm_ieee1905_cmdu_handler;
	ret = ubus_register_event_handler(p->ctx, &p->ieee1905_cmdu, "ieee1905.cmdu");
	if (ret) {
		fprintf(stderr, "delm_register_event() failed %s for ieee1905.cmdu\n",
			ubus_strerror(ret));
	}

	if (!p->cfg.is_collector) {
		p->ieee1905_ev.cb = delm_ieee1905_event_handler;
		ret = ubus_register_event_handler(p->ctx, &p->ieee1905_ev, "ieee1905");
		if (ret) {
			fprintf(stderr, "delm_register_event() failed %s for ieee1905\n",
					ubus_strerror(ret));
		}
	}

	if (p->cfg.is_collector) {
		p->topology_ev.cb = delm_topology_event_handler;
		ret = ubus_register_event_handler(p->ctx, &p->topology_ev, "ieee1905.neighbor.*");
		if (ret) {
			fprintf(stderr, "delm_register_event() failed %s\n for ieee1905.neighbor",
					ubus_strerror(ret));
		}
	}

	if (!p->cfg.is_collector) {
		p->sta_ev.cb = delm_sta_event_handler;
		ret = ubus_register_event_handler(p->ctx, &p->sta_ev, "wifi.sta");
		if (ret) {
			fprintf(stderr, "delm_register_event() failed %s\n for wifi.sta",
					ubus_strerror(ret));
		}
	}

	return ret;
}

uint32_t delm_lookup_object(struct delm_private *p, const char *objname)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(p->ctx, objname, &id);
	if (status != UBUS_STATUS_OK) {
		fprintf(stderr, "object '%s' not present!\n", objname);
		return OBJECT_INVALID;
	}

	return id;
}

static void ieee1905_info_response_cb(struct ubus_request *req,
					int mtype, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *intf_arr;
	struct json_object *intf_obj;
	struct json_object *tmp;
	struct json_object *mac;
	struct delm_private *p;
	uint32_t num = 0;
	uint32_t i;
	char *addr = NULL;
	char *str;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s:Message recieved is NULL\n", __func__);
		return;
	}

	p = (struct delm_private *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "interface_num", &tmp))
		num = json_object_get_int(tmp);

	if (json_object_object_get_ex(jobj, "interfaces", &intf_arr)) {
		for (i = 0; i < num; i++) {
			intf_obj = json_object_array_get_idx(intf_arr, i);
			strncpy(p->iifs.ifname[p->iifs.num],
				json_object_get_string(intf_obj), 15);
			fprintf(stderr, "%s: %d: ieee1905 iface[%d] = %s\n",
					__func__, __LINE__,
					p->iifs.num,
					p->iifs.ifname[p->iifs.num]);
			p->iifs.num++;
		}
	}

	/* Store self device mac addr */
	if (json_object_object_get_ex(jobj, "ieee1905id", &mac)) {
		addr = strdup(json_object_get_string(mac));
		hwaddr_aton(addr, p->ieee1905_macaddr);
		free(addr);
	}

	json_object_put(jobj);
}

int delm_get_ieee1905_interfaces(struct delm_private *p)
{
	struct blob_buf bb = { 0 };
	int ret = 0;

	/* clear out existing ifnames */
	memset(&p->iifs, 0, sizeof(p->iifs));	//TODO: revisit

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(p->ctx, p->ieee1905, "info", bb.head,
				ieee1905_info_response_cb, p, 1000);
	if (ret)
		fprintf(stderr, "Failed to get ieee1905 info (err = %s)\n",
						ubus_strerror(ret));

	blob_buf_free(&bb);

	return ret;
}

static void topology_nodes_response_cb(struct ubus_request *req,
				       int mtype, struct blob_attr *msg)
{
	struct json_object *nbr_array, *tmp, *mac, *self;
	struct json_object *jobj = NULL;
	uint8_t self_node_mac[6] = {0};
	int total_nodes = 0;
	char *str = NULL;
	int num;
	int i;
	struct map_devices {
		int num;
		uint8_t *dev;
	} *narray = NULL;


	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	narray = (struct map_devices *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;


	/* Get own 1905 macaddress */
	if (json_object_object_get_ex(jobj, "self", &self)) {
		if (json_object_object_get_ex(self, "ieee1905_macaddr", &mac)) {
			char *addr = strdup(json_object_get_string(mac));
			uint8_t macaddr[6] = { 0 };

			fprintf(stderr, "Device Self Addr:|%s|\n", addr);

			hwaddr_aton(addr, macaddr);
			memcpy(self_node_mac, macaddr, 6);
			total_nodes += 1;
			free(addr);
		}
	}

	if (json_object_object_get_ex(jobj, "nodes", &nbr_array)) {
		num = json_object_array_length(nbr_array);
		if (!num) {
			fprintf(stderr, "%s: nodes array is empty\n", __func__);
		}

		total_nodes += num;

		narray->dev = calloc(total_nodes, 6 * sizeof(uint8_t));
		if (!narray->dev) {
			fprintf(stderr, "Failed to malloc nodes array\n");
			goto out_err;
		}

		narray->num = total_nodes;
		memcpy(narray->dev, self_node_mac, 6);

		for (i = 0; i < num; i++) {
			char *addr = NULL;

			tmp = json_object_array_get_idx(nbr_array, i);
			if (json_object_object_get_ex(tmp, "ieee1905_macaddr", &mac)) {
				uint8_t tmpmac[6] = { 0 };

				addr = strdup(json_object_get_string(mac));
				fprintf(stderr, "Neighbor Device Addr:|%d:[%s]|\n",
					i, addr);

				hwaddr_aton(addr, tmpmac);
				memcpy(&narray->dev[6 * (i + 1)], tmpmac, 6);
				free(addr);
			}
		}
	}


out_err:
	json_object_put(jobj);
}

static void topology_status_response_cb(struct ubus_request *req,
					int mtype, struct blob_attr *msg)
{
	struct json_object *jobj = NULL;
	struct json_object *tmp;
	char *str;
	int *is_dump_available = NULL;

	if (!msg || !req->priv) {
		fprintf(stderr, "%s: msg = NULL\n", __func__);
		return;
	}

	is_dump_available = (int *)req->priv;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jobj = json_tokener_parse(str);
		free(str);
	}

	if (jobj == NULL)
		return;

	if (json_object_object_get_ex(jobj, "status", &tmp)) {
		str = strdup(json_object_get_string(tmp));

		if (str && !strcmp(str, "available")) {
			*is_dump_available = 1;
		}
		free(str);
	}

	json_object_put(jobj);
}

int delm_get_deagents(struct delm_private *p, int *num, uint8_t **macaddrs)
{
	int is_available = 0;
	struct blob_buf bb;
	int ret = 0;
	struct map_devices {
		int num;
		uint8_t *dev;
	} mapdevs = { 0 };


	/* keep trying until topology is available */
	while (!is_available) {
		memset(&bb, 0, sizeof(struct blob_buf));
		blob_buf_init(&bb, 0);
		ret = ubus_invoke(p->ctx, p->topology, "status", bb.head,
				  topology_status_response_cb,
				  (void *)&is_available, 1000);
		if (ret || (!ret && !is_available)) {
			err("Error getting 'topology'. Retry after 2s\n");
			sleep(2);

			/* update topology object */
			if (ret == UBUS_STATUS_NOT_FOUND) {
				info("update 'topology' object\n");
				p->topology = delm_lookup_object(p, "topology");
				if (p->topology == OBJECT_INVALID) {
					err("Failed to get 'topology' object!\n");
				}

			}
		}

		blob_buf_free(&bb);
	}

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	ret = ubus_invoke(p->ctx, p->topology, "dump", bb.head,
			  topology_nodes_response_cb, &mapdevs, 1000);
	if (!ret) {
		*num = mapdevs.num;
		*macaddrs = mapdevs.dev;
		fprintf(stderr, "num ieee1905 nodes = %d\n", *num);
	} else {
		fprintf(stderr, "Failed to get 'topology' dump\n");
	}

	blob_buf_free(&bb);

	return ret;
}

int self_node_query(struct delm_private *p)
{
	struct blob_buf bb;
	int ret = 1;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	ret = ubus_invoke(p->ctx, p->ieee1905, "query_topology", bb.head,
			NULL, NULL, 1000);
	if (ret) {
		fprintf(stderr, "Failed to get 'ieee1905' query_topology\n");
	}

	blob_buf_free(&bb);

	return ret;
}
