/*
 * main.c
 * WiFi Data Elements main (for standalone build).
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#ifndef WFA_DELM_STANDALONE
#include <map-agent/mapagent.h>
#else
#include "timer.h"
#endif

#include <map1905/map2.h>
#include "wfa_delm.h"
#include "delm.h"
#include "debug.h"

bool syslogging = true;

extern int delm_init(void **priv, void *cfg);
extern int delm_exit(void *priv);

void do_daemonize(const char *pidfile)
{
	char buf[128] = {0};
	int flags;
	int f;

	daemon(0, 0);

	if (!pidfile)
		return;

	f = open(pidfile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
	if (f) {
		flags = fcntl(f, F_GETFD);
		if (flags != -1) {
			flags |= FD_CLOEXEC;
			fcntl(f, F_SETFD, flags);
		}
		if (lockf(f, F_TLOCK, 0) < 0) {
			fprintf(stderr, "File '%s' exists. Aborting...\n",
				pidfile);
			exit(-1);
		}
		ftruncate(f, 0);
		snprintf(buf, sizeof(buf), "%ld\n", (long)getpid());
		write(f, buf, strlen(buf));
	}
}

static void usage(void)
{
	fprintf(stderr, "Usage: wfadatad [options]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "   -h               usage\n");
	fprintf(stderr, "   -s <socket path> ubus socket\n");
	fprintf(stderr, "   -t <seconds>,    collection interval in seconds\n");
	fprintf(stderr, "   -C               enable DE Collector mode\n");
	fprintf(stderr, "   -d,              debug mode; i.e. don't daemonize\n");
	fprintf(stderr, "   -p <pidfile>     pid file path\n");
	fprintf(stderr, "   -r               MAP Profile; 1: profile-1, 2: profile-2\n");
	fprintf(stderr, "   -m               enable single AP mode\n");
	fprintf(stderr, "\n");
}

int main(int argc, char **argv)
{
	const char *ubus_socket = NULL;
	struct ubus_context *ubus_ctx;
	struct delm_config cfg = {0};
	const char *pidfile = NULL;
	bool daemonize = true;
	void *delm_handle;
	int ret;
	int ch;

	while ((ch = getopt(argc, argv, "Cht:dp:s:r:m")) != -1) {
		switch (ch) {
		case 'h':
			usage();
			exit(0);
		case 's':
			ubus_socket = optarg;
			break;
		case 't':
			cfg.refresh_int = atoi(optarg);
			cfg.refresh_int *= 1000;	/* to msecs */
			break;
		case 'C':
			cfg.is_collector = true;
			break;
		case 'd':
			daemonize = false;
			break;
		case 'p':
			pidfile = optarg;
			break;
		case 'r':
			cfg.profile = atoi(optarg);
			break;
		case 'm':
			cfg.is_sap_mode = true;
			break;
		default:
			break;
		}
	}

	if (cfg.is_sap_mode && cfg.is_collector) {
		fprintf(stderr, "collector can not start in single-AP mode\n");
		usage();
		exit(0);
	}

	/* FIXME: Hardcoded, agent profile to profile-2 in case of
	 * invalid profile/no profile.
	 * Hardcoded, collector profile to profile-2.
	 */
	if (!cfg.is_collector) {
		if ((cfg.profile != 1) && (cfg.profile != 2))
			cfg.profile = MAP_PROFILE_2;
	} else if (cfg.is_collector)
		cfg.profile = MAP_PROFILE_2;

	argc -= optind;
	argv += optind;

	if (!pidfile)
		pidfile = cfg.is_collector ?
			"/var/run/wfadatad-collector.pid" :
			"/var/run/wfadatad.pid";

	if (daemonize)
		do_daemonize(pidfile);

	uloop_init();
	ubus_ctx = ubus_connect(ubus_socket);
	if (!ubus_ctx) {
		fprintf(stderr, "Failed to connect to ubus\n");
		return -1;
	}
	ubus_add_uloop(ubus_ctx);
	cfg.uctx = ubus_ctx;

	start_logging("wfadatad");
	ret = delm_init(&delm_handle, &cfg);
	if (ret) {
		fprintf(stderr, "wfa_delm: Failed to start.\n");
		goto out_exit;
	}

	uloop_run();

out_exit:
	stop_logging();
	uloop_done();
	delm_exit(delm_handle);
	return 0;
}
